import mqtt, {MqttClient} from "mqtt";
import getMockMQTT from "./getMockMQTT";
import {Changes, Signal} from "../proto/cdplogreader";
import faker from "@faker-js/faker";
import {testVariables} from "./testConst";

/**
 * The mocked vessel client singleton
 */
let client: MqttClient

/**
 * Creates and connects an MQTT client to the mock broker.
 * Looks for all requests towards vessels.
 * On request it generates some dummy data based on the request and returns it
 */
export const start = async (optClient?: mqtt.MqttClient) => {
    if (client == null) {
        client = optClient ?? await getMockMQTT()
    }
    client.subscribe("request/#")
    client.on("message", ((topic, payload) => {
        const topicParts = topic.split("/")
        if (topicParts.length != 4) return
        if (topicParts[0] != "request") return;
        if (topicParts[2] != "cdplogreader") return;
        if (topicParts[3] != "changes") return;
        const vessel = topicParts[1]
        const cs = Changes.decode(payload)
        let signals: Signal[] = []

        const changes: Changes = {
            fromTo: cs.fromTo,
            lastN: cs.lastN,
            signal: signals
        }

        if (cs.fromTo != null) { // Create getrange response data
            const from = parseInt(cs.fromTo?.fromTime!)
            const to = parseInt(cs.fromTo?.toTime!)
            if (from > to) return;

            let cur = from
            const names = cs.signal.map(s => s.name)
            names.forEach(name => {
                signals.push({
                    name: name,
                    double: {}
                })
            })

            // Randomize "random" and sine data
            const sine = faker.datatype.boolean()
            const step = 25
            if (sine) {
                const sineWidth = faker.datatype.number({min: 1000, max: 10000})
                while (cur < to) {
                    names.forEach((name, idx) => {
                        signals[idx].double[cur+""] = Math.sin(cur*(Math.PI/sineWidth))*100
                    })
                    cur += step
                }
            } else {
                while (cur < to) {
                    names.forEach((name, idx) => {
                        signals[idx].double[cur+""] = faker.datatype.number({min:0, max:200, precision: 0.1})-100
                    })
                    cur += step
                }
            }

        } else if (cs.lastN != null && cs.signal.length != 0) { // Create lastn dummy data
            const last = parseInt(cs.lastN.nChanges)
            const names = cs.signal.map(s => s.name)
            names.forEach(name => {
                signals.push({
                    name: name,
                    double: {}
                })
            })
            const now = Date.now()
            for (let i = 0; i < last; i++) {
                names.forEach((name,idx) => {
                    signals[idx].double[(now-i*10)+""] = faker.datatype.number({precision: 0.001})
                })
            }
        } else if (cs.lastN?.nChanges == "1" && cs.signal.length == 0) { // Create discover dummy data
            const names = testVariables
            names.forEach(name => {
                signals.push({
                    name: name,
                    double: {}
                })
            })
        }

        changes.signal = signals

        client.publish(`response/${vessel}/cdplogreader/changes`, Buffer.from(Changes.encode(changes).finish()))
    }))
    return
}

export const stop = () => {
    client.end(true)
}