import { Command, CommandResponse } from '../../proto/wlCommand'
import Reading from '../../models/reading'
import { InferAttributes, Op, WhereOptions } from 'sequelize'
import EntityDataController from '../../services/database/entityDataController'
import MetaData from '../../models/metaData'

/**
 * Wipes the cache based on the given parameters
 *
 * @param {Command} command wipe command instance
 * @returns {Promise<CommandResponse>} The response to the request
 */
const wipeCache = async (command: Command): Promise<CommandResponse> => {
  const query: WhereOptions<InferAttributes<Reading>> = {}
  const metadataQuery: WhereOptions<InferAttributes<MetaData>> = {}
  let partialQuery: WhereOptions<InferAttributes<MetaData>> = {}
  const ranges = {
    from: command.ranges?.from != null ? parseInt(command.ranges.from) : undefined,
    to: command.ranges?.to != null ? parseInt(command.ranges.to) : undefined
  }

  // If variables are given, build the query to only affect the given variables
  if (command.variables.length !== 0) {
    query.variable = metadataQuery.variable = {
      [Op.in]: await Promise.all(command.variables.map(async (variable) => {
        const node = await EntityDataController.variableByName(variable)
        return node.id
      }))
    }
  }

  // If a vessel is given, build the query to only affect the given vessel
  if (command.vessel != null) {
    const node = await EntityDataController.vesselByName(command.vessel)
    query.vessel = metadataQuery.vessel = node.id
  }

  // If any range data is given, build the query to only affect the correct data by time
  if (ranges != null) {
    if (ranges.to != null && ranges.from != null) {
      partialQuery = {
        ...metadataQuery,
        [Op.or]: [
          {
            from: { [Op.lt]: ranges.to },
            to: { [Op.gt]: ranges.to }
          },
          {
            from: { [Op.lt]: ranges.from },
            to: { [Op.gt]: ranges.from }
          }
        ]
      }

      query.timestamp = metadataQuery.from = metadataQuery.to = {
        [Op.and]: [
          { [Op.gte]: ranges.from },
          { [Op.lte]: ranges.to }
        ]
      }
    } else if (ranges.from != null) {
      partialQuery = {
        ...metadataQuery,
        from: { [Op.lt]: ranges.from },
        to: { [Op.gt]: ranges.from }
      }
      query.timestamp = metadataQuery.from = { [Op.gte]: ranges.from }
    } else if (ranges.to != null) {
      partialQuery = {
        ...metadataQuery,
        from: { [Op.lt]: ranges.to },
        to: { [Op.gt]: ranges.to }
      }
      query.timestamp = metadataQuery.to = { [Op.lte]: ranges.to }
    }
  }
  try {
    if (Object.keys(query).length !== 0) { // If a query has been built, use it
      const transaction = await MetaData.sequelize!.transaction({ logging: false })

      const partials = await MetaData.findAll({ where: partialQuery, logging: false, transaction })

      for (let i = 0; i < partials.length; i++) {
        const partial = partials[i]

        if (ranges.from != null && ranges.to != null && partial.from < ranges.from && partial.to > ranges.to) {
          await MetaData.create({
            variable: partial.variable,
            vessel: partial.vessel,
            from: ranges.to,
            to: partial.to
          }, { transaction, logging: false })
          await partial.update({ to: ranges.from }, { transaction, logging: false })
        } else if (ranges.from != null && partial.from < ranges.from && partial.to > ranges.from) {
          await partial.update({ to: ranges.from }, { transaction, logging: false })
        } else if (ranges.to != null && partial.from < ranges.to && partial.to > ranges.to) {
          await partial.update({ from: ranges.to }, { transaction, logging: false })
        } else {
          throw new Error('This should not be here')
        }
      }

      await Reading.destroy({ where: query, force: true, logging: false, transaction })
      await MetaData.destroy({ where: metadataQuery, force: true, logging: false, transaction })
      await transaction.commit()
    } else { // If no query has been built, perform a full wipe
      await Reading.sequelize?.sync({ force: true, logging: false })
    }
    return { success: true }
  } catch (e: any) { // If any error occurs, forward it in the response
    console.log(e)
    return { success: false, error: e.message }
  }
}
export default wipeCache
