import { Changes } from '../../proto/cdplogreader'
import writeInLog from '../../services/logging/writeInLog'
import { ByteCounter } from '../../services/logging/byteCounter'
import MqttPool from '../../services/mqtt/mqttPool'

/**
 * Fetches the names of a vessels data
 *
 * @param {string} vessel The fessel to retreive data from
 * @param {ByteCounter} byteCounter the {@see ByteCounter} to accumulate data cost to
 * @returns {Promise<string[]>} A list of the logged variables at a given vessel
 */
const discover = async (vessel: string, byteCounter: ByteCounter): Promise<string[]> => {
  const pub = `request/${vessel}/cdplogreader/changes`
  const sub = `response/${vessel}/cdplogreader/changes`

  return await new Promise<string[]>(async (resolve, reject) => {
    // Create the discover request
    const changes: Changes = {
      fromTo: undefined,
      lastN: { nChanges: '1' },
      signal: []
    }

    // Create and set up mqtt client to handle the response
    const mqtt = await MqttPool.get()
    mqtt.subscribe(sub, { qos: 1 })

    // Add response handler
    mqtt.on('message', (topic, payload) => {
      try {
        const payLoadLength = payload.length
        const data = Changes.decode(payload)
        if (data.lastN?.nChanges === '1') {
          clearTimeout(timeout)

          writeInLog(payLoadLength)
          byteCounter.add(payLoadLength)

          // Turn signals into array of variable names
          const dataTypes = data.signal.reduce<string[]>((acc, cur) => {
            return acc.concat(cur.name)
          }, [])
          resolve(dataTypes)
        }
      } catch (e) {
        reject(e)
      } finally {
        MqttPool.release(mqtt)
      }
    })

    // If no request is retrieved within 10 seconds, abort
    const timeout = setTimeout(() => {
      reject('10s timeout')
    }, 1e4)

    // Publish request
    mqtt.publish(pub, Buffer.from(Changes.encode(changes).finish()), { qos: 1 })
  })
}
export default discover
