import { Changes, Signal } from '../../proto/cdplogreader'
import writeInLog from '../../services/logging/writeInLog'
import { ByteCounter } from '../../services/logging/byteCounter'
import MqttPool from '../../services/mqtt/mqttPool'

/**
 * Fetches the last n (lastn) changes for a given vessel and data
 *
 * @param {string} vessel The vessel to retreive data from
 * @param {string[]} variables The variables to retreive
 * @param {number} lastn The count of data to retreive
 * @param {ByteCounter} byteCounter The bytecounter to add accumulated datacost to
 * @returns {Promise<Signal[]>} the last n logged datapoints from each variable as {@see Signal}
 */
const fetchLast = (vessel: string, variables: string[], lastn: number, byteCounter?: ByteCounter): Promise<Signal[]> => {
  const pub = `request/${vessel}/cdplogreader/changes`
  const sub = `response/${vessel}/cdplogreader/changes`

  return new Promise<Signal[]>(async (resolve, reject) => {
    const changes = {
      lastN: {
        nChanges: lastn.toString()
      },
      fromTo: undefined,
      signal: variables.map(d => {
        return {
          name: d,
          double: {}
        }
      })
    }

    // Create mqtt client and set it up to handle the response
    const mqtt = await MqttPool.get()
    mqtt.subscribe(sub, { qos: 1 })
    mqtt.on('message', async (topic, payload) => {
      try {
        const changeData = Changes.decode(payload)
        if (changeData.lastN?.nChanges !== changes?.lastN?.nChanges) return
        const signals = changeData.signal

        const names: string[] = signals.map(s => s.name)
        if (variables.filter(d => !names.includes(d)).length !== 0) return
        clearTimeout(timeout)

        byteCounter?.add(payload.length)
        writeInLog(payload.length)

        resolve(signals)
      } catch (e) {
        reject(e)
      } finally { // Always end the client
        MqttPool.release(mqtt)
      }
    })

    // Publish request, abort if no response is received in 10 seconds
    mqtt.publish(pub, Buffer.from(Changes.encode(changes).finish()), { qos: 1 })
    const timeout = setTimeout(() => reject('10s timeout'), 1e4)
  })
}

export default fetchLast
