import fs from 'fs'
import * as readline from 'readline'
import { StreamStateEntry } from '../../../services/database/streamStateController'

/**
 * Fetches the "double" data from a csv file
 * The data will be formated as {timestamp: value}, so vessel and variable contexsts must be maintained elsewhere
 *
 * @param {StreamStateEntry} entry the files streamstate entry
 * @param {number} from the unix timestamp to start loading from
 * @param {number} to the unix timestamp to load until
 * @returns {Promise<Record<number, number>>} a promise with the timestamp-value entries
 */
export const fetchFromFile = async (entry: StreamStateEntry, from: number, to: number): Promise<Record<number, number>> => {
  // Add use to entry so the file won't be deleted during the read duration
  entry.addUse()

  try {
    if (from > to) throw new Error('From cannot be greater than to')

    // Wait until the entry is ready
    if (entry.status !== true) {
      const status = await new Promise(resolve => {
        entry.sub((a: string|boolean) => {
          resolve(a)
        })
      })

      if (status !== true) {
        throw new Error('Entry status error')
      }
    }

    const double = await new Promise<{[key: number]: number}>((resolve, reject) => {
      const double: {[key: number]: number} = {}

      if (!fs.existsSync(entry.localPath!)) {
        return reject(new Error('File does not exist'))
      }

      const reader = readline.createInterface({
        input: fs.createReadStream(entry.localPath!)
      })

      reader.on('line', (chunk) => {
        const parts = chunk.split(',')
        const timestamp = parseInt(parts[1])
        if (timestamp >= from) {
          if (timestamp <= to) {
            const value = parseFloat(parts[0])
            double[timestamp] = value
          } else {
            reader.close()
          }
        }
      })

      // Once the reader has reached the end, or been closed manually, return the data
      reader.on('close', () => {
        resolve(double)
      })
    }).catch(e => {
      throw e
    })
    return double
  } finally {
    // Always subtract a usage from entry, or it will not be removed
    entry.subUse()
  }
}
export default fetchFromFile
