import { Changes, Signal } from '../../../proto/cdplogreader'
import { MqttClient } from 'mqtt'
import writeInLog from '../../../services/logging/writeInLog'
import { MetaDataAggregator } from '../../../services/database/metaDataAggregator'
import MetaData from '../../../models/metaData'
import { ByteCounter } from '../../../services/logging/byteCounter'
import { InferCreationAttributes } from 'sequelize'
import fs from 'fs'
import path from 'path'
import dataStreamer from '../../../services/database/dataStreamer'
import { EntityData } from '../../../services/database/entityDataController'
import { performance } from 'perf_hooks'
import StreamStateController, { StreamStateEntry } from '../../../services/database/streamStateController'
import MqttPool from '../../../services/mqtt/mqttPool'

/**
 * Interface for optionally using fetchrange within a larger request
 */
export interface FetchRangeMeta {
    client?: MqttClient,
    attempt?: number,
    metadata?: boolean,
    byteCounter?: ByteCounter
}

/**
 * The interval length to start with, in milliseconds
 */
const INITIAL_INTERVAL = 6e4

/**
 * Fetches data within a given range in unix timestamps on a given vessel
 *
 * @param {EntityData} vessel The entitydata representing the vessel
 * @param {EntityData} variable The entitydate representing the variable
 * @param {{from: number, to: number}} range The range to request
 * @param {number} range.from The start of the range to request
 * @param {number} range.to The end of the range to request
 * @param {FetchRangeMeta} meta Optional property to use fetchrange in a retry loop and/or to inject an mqtt client
 * @returns {Promise<Signal>} The fetched data
 */
const fetchRange = async (vessel: EntityData,
  variable: EntityData,
  range: {from: number, to: number},
  meta: FetchRangeMeta = { metadata: true }): Promise<Signal> => {
  const pub = `request/${vessel.name}/cdplogreader/changes`
  const sub = `response/${vessel.name}/cdplogreader/changes`

  return new Promise<Signal>(async (resolve, reject) => {
    if (range.from >= range.to) {
      resolve({ name: variable.name, double: {} })
      return
    }

    // Create streamstate entry to avoid double fetches
    let streamStateEntry: StreamStateEntry
    try {
      streamStateEntry = StreamStateController.create(vessel.name, variable.name, range.from, range.to)
    } catch (e) {
      // If an entry already exists, drop the range-fetch
      return resolve({
        name: variable.name,
        double: {}
      })
    }

    // Create temporary csv file to store fetched records
    let tmpPath = path.resolve(process.cwd(), './tmp/' + Date.now() + '.csv')
    if (!fs.existsSync(path.resolve(process.cwd(), './tmp/'))) fs.mkdirSync(path.resolve(process.cwd(), './tmp/'))
    while (fs.existsSync(tmpPath)) tmpPath = path.resolve(process.cwd(), './tmp/' + Date.now() + '.csv')
    const writeStream = fs.createWriteStream(tmpPath)

    streamStateEntry.setPath(tmpPath)

    const changes = {
      lastN: undefined,
      fromTo: {
        fromTime: range.from + '',
        toTime: range.to + ''
      },
      signal: [{
        name: variable.name,
        double: {}
      }]
    }

    const mqtt = meta?.client != null ? meta.client : await MqttPool.get()
    mqtt.subscribe(sub, { qos: 1 })

    let timeout: NodeJS.Timeout
    const metadata: InferCreationAttributes<MetaData, { omit: 'id' }>[] = []

    const signal: Signal = { name: variable.name, double: {} }
    let currentInterval = INITIAL_INTERVAL
    let from = range.from
    let attempt = 1

    // Fetch the whole interval in smaller intervals
    while (from < range.to && attempt < 3) {
      let to = from + currentInterval
      if (to > range.to) to = range.to

      // console.log(Math.trunc(((from - range.from) / (range.to - range.from)) * 100) + '%')

      const start = performance.now()
      try {
        const part = await new Promise<Signal>((resolve, reject) => {
          timeout = requestTimeout(reject)

          /**
           * Handles an incoming response
           *
           * @param {string} topic The response topic
           * @param {Buffer} payload The response payload
           */
          const handler = async (topic: string, payload: Buffer) => {
            try {
              const changeData = Changes.decode(payload)
              const signals = changeData.signal
              if (changeData.fromTo?.fromTime !== from.toString() || changeData.fromTo.toTime !== to.toString()) return
              if (signals.length !== 1) return
              if (signals[0].name !== variable.name) return

              clearTimeout(timeout)
              const responseTime = performance.now() - start
              /*
               * 1.25 Since 5*1.4 = 6.25 = middle of wanted range
               * 0.8 since it will fall below the 7.5 second threshold most of the time, unlikely to over-adjust downwards
               */
              if (responseTime < 5e3) currentInterval = Math.trunc(currentInterval * 1.25)
              else if (responseTime > 7.5e3) currentInterval = Math.trunc(currentInterval * 0.8)

              // Loggins simply for data collection, not necessary to do in production
              writeInLog(payload.length)
              meta?.byteCounter?.add(payload.length)

              if (meta.metadata !== false) {
                metadata.push({
                  from, to, vessel: vessel.id, variable: variable.id
                })
              }

              mqtt.removeListener('message', handler)
              resolve(signals[0])
            } catch (e) { // If handling request failed, reject with metadata to try again
              reject({
                error: e,
                meta: {
                  attempt: meta.attempt,
                  variable,
                  range,
                  vessel
                }
              })
            }
          }
          // Publish the request
          mqtt.on('message', handler)
          mqtt.publish(pub, Buffer.from(Changes.encode(<Changes>{
            ...changes,
            fromTo: {
              fromTime: from.toString(),
              toTime: to.toString()
            }
          }).finish()), { qos: 1 })
        })

        // Take the data from this request and apply it to the main signals value
        Object.keys(part.double).forEach(key => {
          signal.double[key] = part.double[key]
        })

        // Take a step forward and reset the attempt count on success
        from = to
        attempt = 1
      } catch (e: any) {
        if (attempt >= 3) { // Abort after three attempts
          // Notify the stream states observers
          streamStateEntry.setStatus('Fetch failed')
          streamStateEntry.subUse()
          if (meta.client == null) MqttPool.release(mqtt)
          throw new Error('Fetch failed\nReason:' + e)
        }
        /*
         * 0.2 For extreme measures, if lines are terrible this would mean down to a 2.4 second
         * range per request after 2 misses from the first
         * Given about 24 bytes per record, * 2400 which is the max resolution, that's 57'000b
         * Which is 5.7Kb/s (ignoring time in transfer) which is unlikely to occur
         *
         */
        // Retry with shorter period
        currentInterval = Math.trunc(currentInterval * 0.2)
        attempt++
      }
    }

    // Add uses to relevant nodes so they are available for the metadata aggregator
    vessel.addUse()
    variable.addUse()
    resolve(signal)
    if (meta.client == null) MqttPool.release(mqtt)

    // Once the writestream closes, notify streamstate observers
    writeStream.on('close', () => {
      streamStateEntry.setStatus(true)
    })

    // Write all the data to the streamstate file
    Object.keys(signal.double).forEach(key => {
      writeStream.write(`${signal.double[key]},${key},${vessel.id},${variable.id}\n`)
    })
    writeStream.close()

    try {
      // Stream the fetched records into the database
      await dataStreamer('reading', streamStateEntry, async () => {
        try {
          // Once the data is streamed in, insert and aggregate the metadata
          const savedMeta = await MetaData.bulkCreate(metadata, { ignoreDuplicates: true, logging: false })
          MetaDataAggregator.get().add(savedMeta)
          await MetaDataAggregator.get().waitOnRun()
        } catch (e: any) {
          console.log(e)
        }
      }, (e: any) => {
        console.log(e)
      })
    } catch (e) {
      console.log(e)
    } finally {
      // Subtract uses from nodes regardless of whether the request succeeded or failed
      vessel.subUse()
      variable.subUse()
    }
  })
}
export default fetchRange

/**
 * Starts a 10 second countdown which will reject the given reject function in 10 seconds if not cleared
 *
 * @param {(reason: any) => void} reject The reject function of the parent promise
 * @returns {NodeJS.Timeout} The created timeout
 */
const requestTimeout = (reject: (reason: any) => void): NodeJS.Timeout => {
  return setTimeout(() => {
    console.log('Fetch timeout -------------------------------------------------------------------------------------------------------------')
    reject('10s timeout')
  }, 1e4)
}
