import MetaData from '../../../models/metaData'
import { Op } from 'sequelize'
import fetchRange from './fetchRange'
import { ByteCounter } from '../../../services/logging/byteCounter'
import { Signal } from '../../../proto/cdplogreader'
import fetchFromFile from './fetchFromFile'
import EntityDataController, { EntityData } from '../../../services/database/entityDataController'
import mqtt from 'mqtt'
import StreamStateController, { StreamStateEntry, SubFunc } from '../../../services/database/streamStateController'
import MqttPool from '../../../services/mqtt/mqttPool'

/**
 * Pre-fills the given intervals of data
 *
 * @param {EntityData} vessel The {@see EntityData} containing information about the vessel
 * @param {EntityData} variables The {@see EntityData} constaining information about the variables to prefill
 * @param {object} range Object containing the ranges for the request
 * @param {number} range.from Start of prefill range
 * @param {number} range.to End of prefill range
 * @param {ByteCounter} byteCounter The {@see ByteCounter} to accumulate bytecost to
 * @returns {Promise<{signals: Record<string, Signal>, complete: boolean}>}A promise containing the signals retrieved by name as well as a boolean which is true
 * if the returned signals contain all the information for the given range
 */
const fillUnCachedIntervals = async (vessel: EntityData,
  variables: EntityData[],
  range: {from: number, to: number},
  byteCounter: ByteCounter
): Promise<{
    signals: {[key: string]: Signal}
    complete: boolean
}> => {
  const { from, to } = range

  let client: mqtt.MqttClient|undefined

  try {
    // First find out what data is cached
    const metaData = await MetaData.findAll({
      where: {
        [Op.and]: [
          {
            vessel: vessel.id
          }, {
            variable: {
              [Op.in]: variables.map(variable => variable.id)
            }
          }, {
            [Op.or]: [
              {
                [Op.and]: [
                  {
                    from: {
                      [Op.lt]: to
                    }
                  }, {
                    to: {
                      [Op.gte]: to
                    }
                  }
                ]
              },
              {
                [Op.and]: [
                  {
                    to: {
                      [Op.gt]: from
                    }
                  }, {
                    from: {
                      [Op.lte]: from
                    }
                  }
                ]
              },
              {
                [Op.and]: [
                  {
                    to: {
                      [Op.gte]: to
                    }
                  }, {
                    from: {
                      [Op.lte]: from
                    }
                  }
                ]
              },
              {
                [Op.and]: [
                  {
                    from: {
                      [Op.gte]: from
                    }
                  }, {
                    to: {
                      [Op.lte]: to
                    }
                  }
                ]
              }
            ]
          }

        ]
      },
      logging: false
    })

    // If no metadata exists, all data will be fetched
    const complete = metaData.length === 0

    // Map signals by name
    const signalByName: {[key: string]: Signal[]} = {}
    variables.forEach(variable => {
      signalByName[variable.name] = []
    })

    // Map metadata by id
    const metadataById: {[key: number]: MetaData[]} = {}
    variables.forEach(variable => {
      metadataById[variable.id] = []
    })

    let pendingFetches: Promise<{data: Record<number, number>, metadata: MetaData}>[] = []
    const subFuncs: { variable: string, func: SubFunc }[] = []

    // Check what data can be retrieved from streams-in-progress, add found intervals to metadata
    for (let i = 0; i < variables.length; i++) {
      const variable = variables[i]
      const stateEntries = StreamStateController.getInRange(vessel.name, variable.name, from, to)

      // Subscribe to adds for this variable, if any entries are added, load them in and add a metadata entry
      const subFunc = StreamStateController.subOnAdd(vessel.name, variable.name, async function (entry: StreamStateEntry) {
        if (!(from <= entry.from && entry.from <= to) ||
                    (from <= entry.to && entry.to <= to)) return
        const fetchFrom = entry.from > from ? entry.from : from
        const fetchTo = entry.to < to ? entry.to : to
        const id = variable.id

        // Linear, sorted insertion
        const metadata = new MetaData({ from: fetchFrom, to: fetchTo, variable: variable.id, vessel: vessel.id })
        if (metadataById[id].length === 0) {
          metadataById[id].push(metadata)
        } else {
          let i = 0
          while (i < metadataById[id].length && metadataById[id][i].from < metadata.from) i++
          if (i === metadataById[id].length) {
            metadataById[id].push(metadata)
          } else {
            metadataById[id].splice(i, 0, metadata)
          }
        }

        // Retrieve the data from the file
        await handleFetchFromFile(vessel, variable, entry, fetchFrom, fetchTo).then(res => {
          signalByName[variable.name].push({
            name: variable.name,
            double: res.data
          })
        })
      } as Partial<SubFunc>)
      subFuncs.push({ variable: variable.name, func: subFunc })

      // For each streamstate entry relevant for this request, fetch the data
      for (let i1 = 0; i1 < stateEntries.length; i1++) {
        const entry = stateEntries[i1]

        // Find the innermost of both sides of the range
        const relevantFrom = from > entry.from ? from : entry.from
        const relevantTo = to > entry.to ? entry.to : to
        try {
          const res = await handleFetchFromFile(vessel, variable, entry, relevantFrom, relevantTo)
          signalByName[variable.name].push({
            name: variable.name,
            double: res.data
          })
          metaData.push(res.metadata)
        } catch (e) {
          console.warn('Fetch from file failed\n', e)
        } finally {
          entry?.subUse()
        }
      }
    }

    client = await MqttPool.get()

    // Sort the metadata from the database and the streams
    metaData.sort((a, b) => {
      if (a.from < b.from) return -1
      else if (a.from > b.from) return 1
      else return 0
    })

    metaData.forEach(meta => {
      metadataById[meta.variable].push(meta)
    })

    await Promise.all(pendingFetches)
    pendingFetches = []
    subFuncs.forEach(sub => {
      StreamStateController.unsubOnAdd(vessel.name, sub.variable, sub.func)
    })

    // Fetch all intervals not covered by metadata from the vessel
    let i = 0
    const length = Object.keys(metadataById).length
    while (i < length) {
      const id = parseInt(Object.keys(metadataById)[i])
      const variable = EntityDataController.variableById(id)!
      const meta = metadataById[id]

      // Await any pending file-fetches to avoid metadata collision
      await Promise.all(pendingFetches)
      pendingFetches = []

      // If there exists metadata, fetch everything around
      if (meta.length !== 0) {
        // First, deal with the edgecases
        if (meta[0].from > from) {
          const signal = await fetchRange(vessel, variable, { from, to: meta[0].from }, { attempt: 1, client, byteCounter })
          signalByName[signal.name].push(signal)
        }
        if (meta[meta.length - 1].to < to) {
          const signal = await fetchRange(vessel, variable, { from: meta[meta.length - 1].to, to }, { attempt: 1, client, byteCounter })
          signalByName[signal.name].push(signal)
        }

        // Then go through all metadata between and look for missing intervals
        let j = 0
        while (j < meta.length - 2) {
          const a = meta[j]
          const b = meta[j + 1]
          if (a.to < b.from) {
            const signal = await fetchRange(vessel, variable, { from: a.to, to: b.from }, { attempt: 1, client, byteCounter })
            signalByName[signal.name].push(signal)
          }
          j++
        }
      } else {
        // If no metadata exists, fetch the whole range
        const signal = await fetchRange(vessel, variable, { from, to }, { attempt: 1, client, byteCounter })
        signalByName[signal.name].push(signal)
      }
      i++
    }

    // Reduce all fetched pieces of data into an array of signals unique by name
    return {
      signals: Object.keys(signalByName).reduce<{[key: string]: Signal}>((acc, cur) => {
        const signal: Signal = {
          name: cur,
          double: {}
        }
        signalByName[cur].forEach(sbn => {
          Object.keys(sbn.double).forEach(key => {
            signal.double[key] = sbn.double[key]
          })
        })
        acc[signal.name] = signal
        return acc
      }, {}),
      complete
    }
  } finally {
    if (client) {
      MqttPool.release(client)
    }
  }
}
export default fillUnCachedIntervals

/**
 * Handles the fetch from file calls and creates metadata for the fetched range.
 * Will request the range from the database if the fetch from file fails, as it is likely
 * the fetch from file failed due to the entry finishing before the request is handled
 *
 * @param {EntityData} vessel {@see EntityData} containing information about the vessel
 * @param {EntityData} variable {@see EntityData} containing information about the variable
 * @param {StreamStateEntry} entry The {@see StreamStateEntry} used in the fetch from file request
 * @param {number} from The start of the interval to fetch
 * @param {number} to The end of the interval to fetch
 */
const handleFetchFromFile = async (vessel: EntityData, variable: EntityData, entry: StreamStateEntry, from: number, to: number): Promise<{data: Record<number, number>, metadata: MetaData}> => {
  const data: Record<number, number> = await fetchFromFile(entry, from, to)

  // Create metadata to represent this fetch from file
  const metadata = new MetaData({
    vessel: vessel.id,
    variable: variable.id,
    from: entry.from,
    to: entry.to
  })

  return { data, metadata }
}
