import loadRange from '../../../services/database/loadRange'
import { Signal } from '../../../proto/cdplogreader'
import fillUnCachedIntervals from './fillUnCachedIntervals'
import { ByteCounter } from '../../../services/logging/byteCounter'
import EntityDataController, { EntityData } from '../../../services/database/entityDataController'

/**
 * Retrieves a range
 *
 * @param {string} vessel The vessel to retrieve from
 * @param {string[]} variables The data to retrieve
 * @param {{from: number, to: number}}range The range to retrieve
 * @param {number} range.from The start of the range to fetch
 * @param {number} range.to The end of the range to fetch
 * @param {number} maxgap The maximum gap time in milliseconds
 * @param {ByteCounter} byteCounter the {@see ByteCounter} to accumulate data cost to
 * @returns {Promise<Signal[]>} The data {@see Signal} from the given parameters
 */
const getRange = async (vessel: string,
  variables: string[],
  range: {from: number, to: number},
  maxgap: number = 10,
  byteCounter: ByteCounter): Promise<Signal[]> => {
  let vesselnode: EntityData|undefined
  let variableNodes: EntityData[]|undefined = []

  try {
    // Find or create the nodes for the vessel and variables
    vesselnode = await EntityDataController.vesselByName(vessel)
    variableNodes = await Promise.all(variables.map(async (variable) => {
      return await EntityDataController.variableByName(variable)
    }))

    // Fill the requested range and retrieve the relevant data that is fetched
    const { signals: signalsByName, complete } = await fillUnCachedIntervals(vesselnode, variableNodes, range, byteCounter)

    // If the range wasn't empty, retrieve the pre-cached data
    if (!complete) {
      await loadRange(vesselnode, variableNodes, range, signalsByName)
    }

    // Combine all signals of the same variables into single arrays
    const signals = Object.keys(signalsByName).reduce((acc: Signal[], cur) => {
      acc.push(signalsByName[cur])
      return acc
    }, [])

    return signals
  } finally {
    // Notify that the nodes are no longer in use by this request
    vesselnode?.subUse()
    variableNodes?.forEach(node => node.subUse())
  }
}
export default getRange
