import discover from './discover'
import { Changes } from '../../proto/cdplogreader'
import getRange from './getRange/getRange'
import fetchLast from './fetchLast'
import writeOutLog from '../../services/logging/writeOutLog'
import mqtt from 'mqtt'
import byteCounter from '../../services/logging/byteCounter'
import { DATA_RESPONSE_TOPIC } from '../../services/mqtt/topics'
import { Worker } from 'worker_threads'
import path from 'path'
import { NAMEVALIDATOR } from '../../models/vessel'

/**
 *
 */
class InputError extends Error {
  /**
   * Creates new InputError instance
   *
   * @param {string} message The error message
   */
  constructor (message: string) {
    super(message)
    this.name = InputError.getName
    Object.setPrototypeOf(this, new.target.prototype) // restore prototype chain
  }

  static getName = 'InputError'
}

/**
 * Handles incoming request from client, checks parameter validity and calls the appropriate handler
 *
 * @param {string} vessel The vessel in request
 * @param {Changes} changes {@see Changes} the changes object containing the request-data
 * @param {mqtt.MqttClient} mqtt The MQTT client to use to handle request
 */
const handleRequest = async (vessel: string, changes: Changes, mqtt: mqtt.MqttClient) => {
  const maxgap = 5

  if (!vessel) return
  if (!NAMEVALIDATOR.test(vessel)) {
    changes.error = {
      message: 'Invalid vessel',
      code: 400
    }
    mqtt.publish(DATA_RESPONSE_TOPIC(vessel), Buffer.from(Changes.encode(changes).finish()))
    return
  }

  // Validate variables if any are given
  const variables = changes.signal.map(s => s.name)
  if (variables.length !== 0) {
    for (let i = 0; i < variables!.length; i++) {
      if (!NAMEVALIDATOR.test(variables![i])) {
        variables?.splice(i, 1)
        i--
      }
    }
    if (variables?.length === 0) return
  }

  // Counter to accumulate how much data the request will have used
  const datacost = byteCounter()

  try {
    if (variables.length === 0 && changes.lastN?.nChanges === '0') { // Discover request
      const signals = await discover(vessel, datacost)
      changes.signal = signals.map(signal => {
        return {
          name: signal,
          double: {}
        }
      })
    } else if (changes.fromTo != null && variables.length !== 0) { // Range request
      // Validate range-request-specific parameters
      const from = parseInt(changes.fromTo?.fromTime)
      const to = parseInt(changes.fromTo?.toTime)

      // Add 10 second delay to pad for time drift and delay of data availability on vessels
      const now = Date.now()
      if (to > now - 1e4) {
        throw new InputError('Cannot request data within the last 10 seconds')
      } else
      if (to === from) {
        throw new InputError('Interval cannot start and end at the same time')
      } else
      if (to < from) {
        throw new InputError('To must be greater than from')
      } else
      if (to < 0 || from < 0) {
        throw new InputError('Negative ranges')
      } else
      if ((to - from) * variables.length > 3.6e6) {
        throw new InputError('Cannot request more than an hour of data at once')
      }
      // Handle request
      try {
        const signals = await getRange(vessel, variables, { from, to }, maxgap, datacost)
        changes.signal = signals
      } catch (e: any) {
        changes.signal = variables.map(variable => {
          return {
            name: variable,
            double: {}
          }
        })
        changes.error = {
          message: e?.message ?? e,
          code: 500
        }
      }
    } else if (changes.lastN != null && variables.length !== 0) { // Lastn request
      let lastn = parseInt(changes.lastN.nChanges)
      if (lastn < 10) lastn = 10
      if (lastn > 1e5) lastn = 1e5

      const signals = await fetchLast(vessel, variables, lastn, datacost)
      changes.signal = signals
    } else {
      changes.error = { message: 'Not implemented', code: 503 }
    }

    changes.bytecost = datacost.get()

    console.log('Encoding')
    // Serialize the return-data with protobuf
    const buffer = await new Promise<Buffer>(resolve => {
      const w = new Worker(path.resolve(process.cwd(), './dst/services/workers/changesEncodeWorker.js'), { workerData: changes })
      w.on('message', buffer => {
        resolve(buffer)
      })
    })
    console.log('encoded')

    writeOutLog(buffer.length)

    // Publish the response
    mqtt.publish(DATA_RESPONSE_TOPIC(vessel), buffer)
  } catch (e: any) {
    if (e?.name === InputError.getName) {
      changes.error = { message: e.message, code: 400 }
    } else {
      changes.error = { message: e.message != null ? e.message : e, code: 500 }
    }
    // If an error occurs, publish the received changes object with an error
    mqtt.publish(DATA_RESPONSE_TOPIC(vessel), Buffer.from(Changes.encode(changes).finish()))
  }
}
export default handleRequest
