import { MetaDataList, MetaDataRange, MetaDataRequest, MetaDataResponse } from '../../proto/wlMetaData'
import mqtt from 'mqtt'
import EntityDataController, { EntityData } from '../../services/database/entityDataController'
import MetaData from '../../models/metaData'
import { Op } from 'sequelize'
import { METADATA_RESPONSE_TOPIC } from '../../services/mqtt/topics'
import { NAMEVALIDATOR } from '../../models/vessel'

/**
 * Handle get metadata request
 *
 * @param {string} vessel The vessel to request
 * @param {MetaDataRequest} request The {@see MetaDataRequest}
 * @param {mqtt.MqttClient} client the client to respond with
 */
const getMetaData = async (vessel: string, request: MetaDataRequest, client: mqtt.MqttClient) => {
  // Variable names must be included in the request since all names are hashed in the database
  if (request.variable.length === 0) throw new Error('Variables required')

  // Validate input
  if (!NAMEVALIDATOR.test(vessel)) {
    client.publish(METADATA_RESPONSE_TOPIC(vessel),
      Buffer.from(MetaDataResponse.encode({
        error: `Invalid vessel "${vessel}"`,
        metadata: request.variable.map(variable => {
          return {
            variable,
            ranges: []
          }
        })
      }).finish()))
  }
  for (let i = 0; i < request.variable.length; i++) {
    const variable = request.variable[i]
    if (!NAMEVALIDATOR.test(variable)) {
      client.publish(METADATA_RESPONSE_TOPIC(vessel),
        Buffer.from(MetaDataResponse.encode({
          error: `Invalid variable "${variable}"`,
          metadata: request.variable.map(variable => {
            return {
              variable,
              ranges: []
            }
          })
        }).finish()))
    }
  }

  let node: EntityData|undefined
  let variableNodes: EntityData[]|undefined
  try {
    // Fetch the needed nodes
    node = await EntityDataController.vesselByName(vessel)
    variableNodes = await Promise.all(request.variable.map(variable => {
      return EntityDataController.variableByName(variable)
    }))

    // Build the SQL query based on parameters
    const where: any = {
      vessel: node.id,
      variable: { [Op.in]: variableNodes.map(node => node.id) }
    }
    if (request.from != null && request.to != null) {
      where[Op.or] = [
        {
          from: {
            [Op.and]: [
              { [Op.gte]: request.from },
              { [Op.lte]: request.to }
            ]
          }
        },
        {
          to: {
            [Op.and]: [
              { [Op.gte]: request.from },
              { [Op.lte]: request.to }
            ]
          }
        }
      ]
    } else if (request.from != null) {
      where.from = { [Op.gte]: request.from }
    } else if (request.to != null) {
      where.to = { [Op.lte]: request.to }
    }

    // Load the data
    const metaData = await MetaData.findAll({
      where,
      logging: false
    })

    // Map the metadata by id
    const metadataById: Record<number, MetaData[]> = {}
    metaData.forEach(meta => {
      if (metadataById[meta.variable] == null) metadataById[meta.variable] = []
      metadataById[meta.variable].push(meta)
    })

    // Then map it by name using the nodes
    const metadataByName = Object.keys(metadataById).reduce<Record<string, MetaData[]>>((acc, id) => {
      const node = EntityDataController.variableById(parseInt(id))
      node.subUse()
      const name = node.name
      acc[name] = metadataById[parseInt(id)]
      return acc
    }, {})

    // Turn the data into metadatalists and create the metadataresponse
    const response: MetaDataResponse = {
      metadata: Object.keys(metadataByName).reduce<MetaDataList[]>((acc, cur) => {
        const record = metadataByName[cur]
        const list: MetaDataRange[] = record.map(meta => {
          return {
            from: meta.from + '',
            to: meta.to + ''
          }
        })
        acc.push({
          variable: cur,
          ranges: list
        })
        return acc
      }, [])
    }

    // Publish the metadata response
    client.publish(METADATA_RESPONSE_TOPIC(vessel), Buffer.from(MetaDataResponse.encode(response).finish()))
  } catch (e: any) {
    console.log(e)
    // If an error occurs, return empty metadata response lists and fill in the error field
    const response: MetaDataResponse = {
      metadata: variableNodes!.map(node => {
        return {
          variable: node.name,
          ranges: []
        }
      }),
      error: e.message ? e.message : e.toString()
    }
    client.publish(METADATA_RESPONSE_TOPIC(vessel), Buffer.from(MetaDataResponse.encode(response).finish()))
  } finally {
    // Notify that this request no longer uses the nodes
    node?.subUse()
    variableNodes?.forEach(node => node.subUse())
  }
}
export default getMetaData
