import configEnv from './configuration/configEnv'
import setupDB from './services/database/setupDB'
import fs, { writeFileSync } from 'fs'
import path from 'path'
import getMQTT from './services/mqtt/getMQTTClient'
import requestTopicHandler from './handlers/data/requestChangesHandler'
import commandTopicHandler from './handlers/commands/commandTopicHandler'
import { Command } from './proto/wlCommand'
import { Changes } from './proto/cdplogreader'
import {
  COMMAND_REQUEST_TOPIC,
  DATA_REQUEST_TOPIC,
  METADATA_REQUEST_TOPIC
} from './services/mqtt/topics'
import getMetaData from './handlers/metadata/getMetaData'
import { MetaDataRequest } from './proto/wlMetaData'
import { Queue } from './services/mqtt/queue'
import { fork } from 'child_process'

/**
 * Loop containing the cache program
 * On MQTT errors, such as connectivity loss, it will restart
 */
;(async () => {
  if (!fs.existsSync(path.resolve(process.cwd(), './logs'))) {
    fs.mkdirSync(path.resolve(process.cwd(), './logs'))
  }
  configEnv()
  try {
    await setupDB()
  } catch (e: any) {
    console.log('Database issue, exiting.')
    process.exit(1)
  }
  const healthCheckClient = fork(path.resolve(process.cwd(), './dst/services/mqtt/healthCheckClient.js'))
  healthCheckClient.on('message', (message) => {
    throw new Error(message.toString())
  })
  try {
    const client = await getMQTT()
    client.on('disconnect', () => {
      client.reconnect()
    })
    client.removeAllListeners('error')
    client.on('error', e => {
      console.log('API client error\n', e.message)
      process.exit()
    })

    const dataRequestQueue = new Queue<{ vessel: string, payload: Buffer }>()

    setInterval(() => {
      if (dataRequestQueue.hasNext()) {
        try {
          const request = dataRequestQueue.dequeue()
          requestTopicHandler(request.vessel, Changes.decode(request.payload), client)
        } catch (e) {
          console.log(e)
        }
      }
    }, 1e2)

    client.subscribe(DATA_REQUEST_TOPIC)
    client.subscribe(COMMAND_REQUEST_TOPIC)
    client.subscribe(METADATA_REQUEST_TOPIC)
    client.on('message', async (topic, payload) => {
      if (payload == null || payload?.length === 0) return
      const topicParts = topic.split('/')
      const requestType = topicParts[1]
      try {
        if (requestType === 'command') {
          console.log('####Client command####')
          commandTopicHandler(Command.decode(payload), client)
        } else if (requestType === 'data') {
          console.log('####Client data request####')
          const vessel = topic.split('/')[2]
          dataRequestQueue.enqueue({ vessel, payload })
        } else if (requestType === 'metadata') {
          console.log('####Client metadata request####')
          const vessel = topic.split('/')[2]
          getMetaData(vessel, MetaDataRequest.decode(payload), client)
        }
      } catch (e: any) {
        console.log('Request failed\nReason:', e.message)
      }
    })
  } catch (e: any) {
    healthCheckClient.send('end')
    console.log('Exiting', '\nReason: ' + (e?.message ?? e))
    writeFileSync(path.resolve(process.cwd(), './logs/breaking'), e.toString())
    process.exit(1)
  }
})()
