import { Pool, PoolClient } from 'pg'
import { from as copyFrom } from 'pg-copy-streams'
import fs from 'fs'
import getConnectionString from './getConnectionString'
import { StreamStateEntry } from './streamStateController'

let pool: Pool

/**
 * Pipes the file referred in the {@see StreamStateEntry} entry into the database through a temp-table
 *
 * @param {string} table The table/relation to pipe into
 * @param {StreamStateEntry} entry The streams entry in the streamstate
 * @param {() => void} onSuccess callback to run on success
 * @param {() => void} onError callback to run on error
 */
const dataStreamer = async (table: string, entry: StreamStateEntry, onSuccess: () => void, onError: (e: any) => void): Promise<void> => {
  let fileStream: fs.ReadStream
  let client: PoolClient

  // If the entry isn't ready, wait for its status to be changed
  if (!entry.status) {
    const result = await new Promise(resolve => {
      entry.sub((a: string|boolean) => resolve(a))
    })

    // If the status is a string, it means it is an error and the stream is aborted
    if (typeof result === 'string') {
      throw new Error(result)
    }
  }
  try {
    await new Promise<void>(async (resolve, reject) => {
    // If no pool exists, create one
      if (pool == null) {
        pool = new Pool({
          connectionString: getConnectionString(),
          connectionTimeoutMillis: 1e4,
          max: 20,
          idleTimeoutMillis: 1e4,
          keepAlive: true,
          ssl: process.env.DB_REQUIRE_TLS !== 'false'
        })
        pool.on('error', async (e) => {
          console.log(e)
        })
      }

      // Connect client and set error listener
      try {
        client = await pool.connect()
        client.on('error', err => {
          console.log('Data streamer pool error', err)
          stream.destroy()
        })
      } catch (e) {
        console.log('Could not connect pool-client')
        reject(e)
        return
      }

      // Start readstream on file
      fileStream = fs.createReadStream(entry.localPath!)

      // Create CopyFrom stream to the temporary table
      const tmpTable = 'tmp_table_' + Date.now()
      const stream = client.query(copyFrom('CREATE TEMP TABLE ' + tmpTable + '\n' +
        'ON COMMIT DROP AS SELECT * FROM reading WITH NO DATA;' +
        'COPY ' + tmpTable + " FROM STDIN WITH DELIMITER ',';" +
        'INSERT INTO reading SELECT * FROM ' + tmpTable + ' ON CONFLICT DO NOTHING;'))

      // If the copystream fails, reject
      stream.on('error', async (e) => {
        console.log('Read-stream error')
      })

      // Resolve once the entire dataset is piped
      stream.on('finish', async () => {
        await onSuccess()
        resolve()
      })
      fileStream.pipe(stream)
    }).catch(async (e) => {
      await onError(e)
    }).finally(() => {
    // If the filestream still exists, destroy it
      if (!(fileStream?.destroyed || fileStream?.readableEnded)) {
        fileStream?.destroy()
      }

      // Notify the entry that it is no longer being used by this handler
      entry?.subUse()

      // Release the client connection from the pool
      try {
        if (client) {
          client.release()
        }
      } catch (e) {
        console.log('Error releasing postgres client')
        console.log(e)
      }
    })
  } catch (e) {
    console.log('Datastream error')
    console.log(e)
  }
}
export default dataStreamer
