/**
 * Class representing the relevant information for a vessel or a avariable
 */
import { Vessel } from '../../models/vessel'
import { createHash } from 'crypto'
import { Variable } from '../../models/variable'

/**
 *
 */
export class EntityData {
  /**
   * A subjects id
   */
  readonly id: number

  /**
   * The sha256 hashed name of a subject
   */
  readonly hash: string

  /**
   * A subjects name
   */
  readonly name: string

  /**
   * The type of subject
   */
  readonly type: 'vessel'|'variable'

  /**
   * The amount of uses relying on this node
   */
  uses : number = 1

  /**
   * Constructs a node automatically set with one use
   *
   * @param {{hash: string, name: string, id: number}} args The internal data for an entitydata node
   * @param {string} args.hash The hash of the nodes name
   * @param {string} args.name The name of the entity the node represents
   * @param {number} args.id The id of the entity the node represents
   * @param {'vessel'|'variable'} type The type of entity the node represents
   */
  constructor (args: {hash: string, name: string, id: number}, type: 'vessel'|'variable') {
    this.id = args.id
    this.hash = args.hash
    this.name = args.name
    this.type = type
  }

  /**
   * Adds a use to the node
   *
   * @throws {Error} If node is already removed
   */
  public addUse () {
    if (this.uses <= 0) {
      throw new Error('Cannot add uses to a removed node')
    }
    this.uses++
  }

  /**
   * Subtracts a use from the node
   * If the node is no longer in use it wil self-destruct
   */
  public subUse () {
    this.uses--
    if (this.uses <= 0) {
      EntityDataController.remove(this)
    }
  }
}

/**
 * The controller managing all entitydata and acts as a repository for currently used data
 */
export default class EntityDataController {
  /**
   * A map over all the vessel entities nodes
   *
   * @private
   */
  private static vesselState: {
        id: Record<number, EntityData>
        name: Record<string, EntityData>
        hash: Record<string, EntityData>
    } = { id: {}, name: {}, hash: {} }

  /**
   * A map over all the vessel entities nodes
   *
   * @private
   */
  private static variableState: {
        id: Record<number, EntityData>
        name: Record<string, EntityData>
        hash: Record<string, EntityData>
    } = { id: {}, name: {}, hash: {} }

  /**
   * Removes a given node from the repository
   *
   * @param {EntityData} node The node to remove
   */
  public static remove (node: EntityData) {
    if (node.type === 'variable') {
      delete EntityDataController.variableState.id[node.id]
      delete EntityDataController.variableState.name[node.name]
      delete EntityDataController.variableState.hash[node.hash]
    } else {
      delete EntityDataController.vesselState.id[node.id]
      delete EntityDataController.vesselState.name[node.name]
      delete EntityDataController.vesselState.hash[node.hash]
    }
  }

  /**
   * Adds a node to the repository
   * If a colliding node exists,the existing node will get an added use and be returned
   *
   * @param {EntityData} node The node to add
   * @returns {EntityData} The added node, this returned value must be used, not the one made
   */
  private static add (node: EntityData): EntityData {
    if (node.type === 'variable') {
      const preCheck = this.variableState.id[node.id]
      if (preCheck == null) {
        this.variableState.id[node.id] = node
        this.variableState.name[node.name] = node
        this.variableState.hash[node.hash] = node
        return node
      } else {
        preCheck.addUse()
        return preCheck
      }
    } else {
      const preCheck = this.vesselState.id[node.id]
      if (preCheck == null) {
        this.vesselState.id[node.id] = node
        this.vesselState.name[node.name] = node
        this.vesselState.hash[node.hash] = node
        return node
      } else {
        preCheck.addUse()
        return preCheck
      }
    }
  }

  /**
   * Returns the vessel mapped to the given id
   *
   * @param {number} id The id of the vessel
   * @returns {EntityData} The found entitydata
   * @throws {Error} If node doesn't exist
   */
  public static vesselById (id: number): EntityData {
    const node = this.vesselState.id[id]
    if (node == null) throw new Error('Vesselnode ' + id + ' does not exist')
    node.addUse()
    return node
  }

  /**
   * Returns the vessel mapped to the given name
   *
   * @param {string} name The name of the vessel
   * @returns {Promise<EntityData>} The found or created entitydata
   */
  public static async vesselByName (name: string): Promise<EntityData> {
    let node: EntityData|undefined = this.vesselState.name[name]
    node?.addUse()
    if (node == null) {
      const hash = createHash('sha256').update(name).digest('hex').toString()
      const id = await Vessel.findOrCreate({ where: { name: hash }, logging: false }).then(res => res[0].id)
      node = new EntityData({
        name, hash, id
      }, 'vessel')
      return this.add(node)
    }
    return node
  }

  /**
   * Returns the vessel mapped to the given hash
   *
   * @param {string} hash The hashed name of the vessel
   * @returns {EntityData} The found entitydata
   * @throws {Error} If node doesn't exist
   */
  public static vesselByHash (hash: string): EntityData {
    const node = this.vesselState.hash[hash]
    node?.addUse()
    if (node == null) throw new Error('Vesselnode ' + hash + ' does not exist')
    return node
  }

  /**
   * Returns the variable mapped to the given id
   *
   * @param {number} id The id of the entitydata
   * @returns {EntityData} The retrieved entitydata
   * @throws {Error} If node doesn't exist
   */
  public static variableById (id: number): EntityData {
    const node = this.variableState.id[id]
    node?.addUse()
    if (node == null) throw new Error('Variablenode ' + id + ' does not exist')
    return node
  }

  /**
   * Returns the variable mapped to the given name
   *
   * @param {string} name The name of the variable
   * @returns {Promise<EntityData>} The found or created entitydata
   */
  public static async variableByName (name: string): Promise<EntityData> {
    let node: EntityData|undefined = this.variableState.name[name]
    node?.addUse()
    if (node == null) {
      const hash = createHash('sha256').update(name).digest('hex').toString()
      const res = await Variable.findOrCreate({ where: { name: hash }, logging: false })
      const id = res[0].id
      node = new EntityData({
        name, hash, id
      }, 'variable')
      return this.add(node)
    }
    return node
  }

  /**
   * Returns the variable mapped to the given hash
   *
   * @param {string} hash The hashed name of the variable
   * @returns {EntityData} The found entitydata
   * @throws {Error} If node doesn't exist
   */
  public static variableByHash (hash: string): EntityData {
    const node = this.variableState.hash[hash]
    node?.addUse()
    if (node == null) throw new Error('Variablenode ' + hash + ' does not exist')
    return node
  }
}
