import { Op } from 'sequelize'
import Reading from '../../models/reading'
import EntityDataController, { EntityData } from './entityDataController'
import { Signal } from '../../proto/cdplogreader'

/**
 * Loads given range from database into the given record
 *
 * @param {EntityData} vessel The {@see EntityData} pertaining to the vessel to query
 * @param {EntityData} variables The {@see EntityData} pertaining to the variables to query
 * @param {{from: number, to: number}} range The range to query
 * @param {number} range.from The start of the range to load
 * @param {number} range.to The end of the range to load
 * @param {Record<string, Signal>} signalsByName The string-to-signal record to load the data into
 */
const loadRange = async (vessel: EntityData, variables: EntityData[], range: {from: number, to: number}, signalsByName: Record<string, Signal>): Promise<number> => {
  try {
    // fetch all records for the given range
    const records: Reading[] = await Reading.findAll({
      attributes: ['variable', 'value', 'timestamp'],
      where: {
        [Op.and]: [
          { vessel: vessel.id },
          {
            timestamp: {
              [Op.and]: [
                { [Op.gte]: range.from },
                { [Op.lte]: range.to }
              ]
            }
          },
          {
            variable: { [Op.in]: variables.map(variable => variable.id) }
          }
        ]
      },
      logging: false
    }) as Reading[]

    // Create an id-reading mapping for all variables
    const cachedById = variables.reduce<Record<number, Reading[]>>((acc, cur) => {
      acc[cur.id] = []
      return acc
    }, {})

    // Add readings to id-reading mapping in 100k increments to avoid blocking healthchecks
    let i = 0
    while (i < records.length) {
      if (i % 100000) {
        await new Promise(resolve => {
          setImmediate(() => resolve(null))
        })
      }
      const reading = records[i]
      cachedById[reading.variable].push(reading)
      i++
    }

    // Take all readings by id, map them to their names, and place them into their respective signalsByName signal
    for (let i = 0; i < Object.keys(cachedById).length; i++) {
      const key = Object.keys(cachedById)[i]
      const id = parseInt(key)
      const node = EntityDataController.variableById(id)
      node.subUse()
      if (signalsByName[node?.name!] == null) signalsByName[node?.name!] = { name: node?.name!, double: {} }
      let j = 0
      while (j < cachedById[id].length) {
        // Surrender execution per 100k records to allow healthcheck to respond
        if (j % 100000) {
          await new Promise(resolve => {
            setImmediate(() => resolve(null))
          })
        }
        const reading = cachedById[id][j]
        signalsByName[node?.name!].double[reading.timestamp] = reading.value
        j++
      }
    }
    return records.length
  } catch (e: any) {
    console.log(e)
    throw new Error('Could not load data from database')
  }
}
export default loadRange
