import MetaData from '../../models/metaData'
import fetchRange from '../../handlers/data/getRange/fetchRange'
import Reading, { ReadingBean } from '../../models/reading'
import { MqttClient } from 'mqtt'
import Sequelize, { Op } from 'sequelize'
import EntityDataController, { EntityData } from './entityDataController'
import MqttPool, { MqttPoolClient } from '../mqtt/mqttPool'

/**
 * Class that aggregates metadata upon requests
 */
export class MetaDataAggregator {
  /**
   * List of metadata accumulated since last aggregation
   *
   * @private
   */
  private metadata: MetaData[] = []

  /**
   * 10 second timeout between aggregations
   *
   * @private
   */
  private timeout: NodeJS.Timeout

  /**
   * Callbacks for promises waiting on run
   *
   * @private
   */
  private waiting: (() => void)[] = []

  /**
   * The interval length which an aggregation will fill if a gap
   * between intervals is smaller
   *
   * @private
   */
  private readonly sensitivity = 1e3

  /**
   * Singleton instance
   *
   * @private
   */
  private static instance: MetaDataAggregator|undefined

  /**
   * Returns singleton instance
   *
   * @returns {MetaDataAggregator} The singleton instance
   */
  public static get = () => {
    if (MetaDataAggregator.instance === undefined) {
      MetaDataAggregator.instance = new MetaDataAggregator()
    }
    return MetaDataAggregator.instance
  }

  /**
   * Private constructor
   *
   * @private
   */
  private constructor () {
    this.timeout = setTimeout(async () => {
      this.run().then(() => MetaDataAggregator.get().resetTimeout())
    })
  }

  /**
   * Sets a new 10-second timeout before running aggregations
   *
   * @private
   */
  private resetTimeout () {
    this.timeout = setTimeout(async () => {
      this.run().then(() => MetaDataAggregator.get().resetTimeout())
    }, 1e4)
  }

  /**
   * Clears the aggregators timeout, in practice ending it
   */
  public kill () {
    clearTimeout(this.timeout)
  }

  /**
   * Function to add one or multiple metadata beans
   *
   * @param {MetaData|MetaData[]} md the metadata to add
   */
  public add = (md: MetaData|MetaData[]): void => {
    if (!Array.isArray(md)) md = [md]
    this.metadata.push(...md)
  }

  /**
   * Awaits on the next time aggregation has run
   */
  public waitOnRun = async (): Promise<void> => {
    await new Promise<void>(resolve => {
      this.waiting.push(resolve)
    })
  }

  /**
   * Loads all preceding data within the set sensitivity of the aggregator
   *
   * @param {EntityData} vessel The entity data node for the vessel
   * @param {EntityData} variable The entity data node for the variable
   * @param {MetaData[]} meta A metadata array containing only entries of the vessel and variable combination
   */
  private async loadPrecedingMetaData (vessel: EntityData, variable: EntityData, meta: MetaData[]) {
    let preceding: MetaData[]
    do {
      preceding = await MetaData.findAll({
        where: {
          [Op.and]: [
            { vessel: vessel.id },
            { variable: variable.id },
            {
              to: {
                [Op.and]: [
                  { [Op.lte]: meta[0].from },
                  { [Op.gte]: meta[0].from - this.sensitivity }
                ]
              }
            }
          ]
        },
        logging: false
      })
      if (preceding.length !== 0) meta.splice(0, 0, ...preceding)
    } while (preceding.length !== 0)
  }

  /**
   * Loads all following data within the set sensitivity of the aggregator
   *
   * @param {EntityData} vessel The entity data node for the vessel
   * @param {EntityData} variable The entity data node for the variable
   * @param {MetaData[]} meta A metadata array containing only entries of the vessel and variable combination
   */
  private async loadFollowingData (vessel: EntityData, variable: EntityData, meta: MetaData[]) {
    let following: MetaData[]
    do {
      following = await MetaData.findAll({
        where: {
          [Op.and]: [
            { vessel: vessel.id },
            { variable: variable.id },
            {
              from: {
                [Op.and]: [
                  { [Op.lte]: meta[meta.length - 1].to + this.sensitivity },
                  { [Op.gte]: meta[meta.length - 1].to }
                ]
              }
            }
          ]
        },
        logging: false
      })
      if (following.length !== 0) meta.splice(meta.length, 0, ...following)
    } while (following.length !== 0)
  }

  /**
   * Loads all data sandwiched within any sub-10 second interval of the currently aggregating ranges
   *
   * @param {EntityData} vessel The entity data node for the vessel
   * @param {EntityData} variable The entity data node for the variable
   * @param {MetaData[]} meta A metadata array containing only entries of the vessel and variable combination
   */
  private async loadIntermediateData (vessel: EntityData, variable: EntityData, meta: MetaData[]) {
    let i = 0
    while (i < meta.length - 1) {
      if (meta[i].to !== meta[i + 1].from && meta[i].to + 1e5 > meta[i + 1].from) {
        const intermediaries = await MetaData.findAll({
          where: {
            [Op.and]: [
              { vessel: vessel.id },
              { variable: variable.id },
              {
                [Op.or]: [
                  { from: { [Op.and]: [{ [Op.gte]: meta[i].to }, { [Op.lt]: meta[i + 1].from }] } },
                  { to: { [Op.and]: [{ [Op.gt]: meta[i].to }, { [Op.lte]: meta[i + 1].from }] } }
                ]
              }
            ]
          },
          order: MetaData.sequelize!.literal('"from" ASC'),
          logging: false
        })
        if (intermediaries.length !== 0) {
          meta.splice(i + 1, 0, ...intermediaries)
          i += intermediaries.length
        } else {
          i++
        }
      } else {
        i++
      }
    }
  }

  /**
   * Asynchronously performs aggregation of all data added since the last aggregation
   *
   */
  public run = async () => {
    const waiting = this.waiting
    this.waiting = []
    const accumulated = this.metadata
    this.metadata = []
    if (accumulated.length === 0) return

    console.log('Aggregating')
    let transaction: Sequelize.Transaction|undefined
    let client: MqttPoolClient|undefined
    // Group metadata by vessels and variables
    const accByVD: {[key: string]: {[key: string]: MetaData[]} } = {}
    accumulated.forEach(acc => {
      if (accByVD[acc.vessel] == null) accByVD[acc.vessel] = {}
      if (accByVD[acc.vessel][acc.variable] == null) accByVD[acc.vessel][acc.variable] = []
      accByVD[acc.vessel][acc.variable].push(acc)
    })

    // Sort data by from-time
    Object.keys(accByVD).forEach(vessel => {
      Object.keys(accByVD[vessel]).forEach(data => {
        accByVD[vessel][data].sort((a, b) => {
          return a.from > b.from ? 1 : a.from === b.from ? 0 : -1
        })
      })
    })

    // Map vessel names and ids
    const vessels = Object.keys(accByVD).map(value => {
      const node = EntityDataController.vesselById(parseInt(value))!
      return node
    })

    // Map variable names and ids
    const variables = Object.keys(accByVD).reduce((acc: string[], cur) => {
      acc.push(...Object.keys(accByVD[cur]))
      return acc
    }, []).map(value => {
      const node = EntityDataController.variableById(parseInt(value))!
      return node
    })

    try {
      transaction = await Promise.resolve(MetaData.sequelize?.transaction({ logging: false })!)

      client = await MqttPool.get()

      // Handle data
      for (const vessel of vessels) {
        for (const variable of variables) {
          const meta = accByVD[vessel.id][variable.id]

          if (meta == null) continue

          // Load any relevant preceding metadata
          await this.loadPrecedingMetaData(vessel, variable, meta)

          // Load any relevant following metadata
          await this.loadFollowingData(vessel, variable, meta)

          // Load all metadata within close intervals
          await this.loadIntermediateData(vessel, variable, meta)

          // Aggregate intervals fulfilling criteria
          let i = 0
          while (i < meta.length - 1) {
            if (meta[i].to + this.sensitivity >= meta[i + 1].from) {
              const updated = await this.aggregate(meta[i], meta[i + 1], client, vessel, variable, transaction)
              if (updated != null) {
                meta.splice(i, 2, updated)
              } else {
                i++
              }
            } else {
              i++
            }
          }
        }
      }
      await Promise.resolve(transaction.commit())

      // Notify all waiting observers that aggregation has been performed
      waiting.forEach(resolve => resolve())
    } catch (e) {
      // Rollback the aggregation on error
      transaction?.rollback()
      console.log(e)
      this.add(accumulated)
      this.waiting.push(...waiting)
    } finally {
      if (client != null) {
        MqttPool.release(client)
      }
      // Always subtract uses from used nodes
      vessels?.forEach(node => {
        node.subUse()
      })
      variables?.forEach(node => {
        node.subUse()
      })
    }
  }

  /**
   * Aggregates two metadata intervals
   *
   * @param {MetaData} a The first metadata entry
   * @param {MetaData} b The second metadata entry
   * @param {MqttClient} client The MQTT client to use to fetch the missing interval
   * @param {EntityData} vessel The entitydata representing the vessel
   * @param {EntityData} variable The entitydata representing the variable
   * @param {Sequelize.Transaction} transaction The transaction to use when aggregating
   */
  private aggregate = async (a: MetaData, b: MetaData, client: MqttClient,
    vessel: EntityData,
    variable: EntityData,
    transaction: Sequelize.Transaction
  ): Promise<MetaData> => {
    if (a.to !== b.from) {
      const mid = (await fetchRange(
        vessel,
        variable,
        { from: a.to, to: b.from },
        { client, attempt: 1, metadata: false })).double
      const records: ReadingBean[] = []

      Object.keys(mid).forEach(d => {
        const record: ReadingBean = {
          variable: a.variable,
          vessel: a.vessel,
          timestamp: parseInt(d),
          value: mid[d]
        }
        records.push(record)
      })

      await Reading.bulkCreate(records, { logging: false, ignoreDuplicates: true })
    }

    await MetaData.destroy({
      where: {
        [Op.or]: [
          { id: b.id },
          {
            [Op.and]: [
              { vessel: a.vessel },
              { variable: a.variable },
              { from: { [Op.gte]: a.to } },
              { to: { [Op.lte]: b.from } }
            ]
          }
        ]
      },
      logging: false,
      transaction
    })
    await MetaData.update({
      to: b.to
    }, {
      where: {
        id: a.id
      },
      logging: false,
      transaction
    })
    a.to = b.to
    return a
  }
}
