import pg from 'pg'
import { Sequelize } from 'sequelize'
import { initReading } from '../../models/reading'
import { initMetaData } from '../../models/metaData'
import { initVessel } from '../../models/vessel'
import { initVariable } from '../../models/variable'
import getConnectionString from './getConnectionString'

/**
 * Initializes the database given in env if it doesn't already exist
 */
const setupDB = async (): Promise<Sequelize> => {
  // Create wirelogger database if it does not exist
  pg.defaults.parseInt8 = true
  const pgClient = new pg.Client({
    connectionString: `postgres://${process.env.DB_USER}` +
            `:${process.env.DB_PWD}` +
            `@${process.env.DB_URL}/postgres`,
    connectionTimeoutMillis: 1e4,
    ssl: process.env.DB_REQUIRE_TLS !== 'false'
  })
  try {
    console.info('Connecting to database...')
    await pgClient.connect()
    console.info('Connected to database')
  } catch (e: any) {
    console.error(`Could not connect to database ${process.env.DB_URL}`, '\nReason: ' + (e.message ?? e), '\nExiting')
    process.exit(1)
  }
  const databases = await pgClient.query(
    'SELECT datname FROM pg_database WHERE datname = \'wirelogger\'')
  if (databases.rows.length === 0) {
    await pgClient.query('CREATE DATABASE wirelogger')
  }
  pgClient.end()

  // Initialize sequelize models
  const connectionString = getConnectionString()
  const sequelize = new Sequelize(connectionString, {
    protocol: 'postgres',
    dialect: 'postgres',
    dialectOptions: {
      ssl: process.env.DB_REQUIRE_TLS !== 'false'
    },
    pool: {
      acquire: 1e4,
      max: 50
    }
  })

  initVariable(sequelize)
  initVessel(sequelize)
  initReading(sequelize)
  initMetaData(sequelize)

  try {
    console.info('Setting up models...')
    await sequelize.authenticate()
    await sequelize.sync({ alter: true, logging: false })
    console.info('Finished setting up models')
  } catch (e: any) {
    console.error('Could not connect to database \nReason:', (e?.message ?? e), '\nExiting')
    process.exit(1)
  }

  // Create TimeScaleDB hypertable
  await sequelize.query('CREATE EXTENSION IF NOT EXISTS timescaledb CASCADE;', { logging: false })
  await sequelize.query('SELECT create_hypertable(\'"reading"\', \'timestamp\', if_not_exists => TRUE, migrate_data => TRUE, chunk_time_interval => 300000);', {
    logging: false
  })
  return sequelize
}
export default setupDB
