import fs from 'fs'
import faker from '@faker-js/faker'

/**
 * Extension of functions with an identifier to be able to identify it in the sub mappings of the streamstate controller
 */
export interface SubFunc {
    (entry: StreamStateEntry): void;
    id: string
}

/**
 * Class defining a streamstate entry
 */
export class StreamStateEntry {
  /**
   * The vessel the entry represents
   */
  public vessel: string

  /**
   * The variable the entry represents
   */
  public variable: string

  /**
   * The start of the entrys interval (in unix time)
   */
  public from: number

  /**
   * The end of the entrys interval (in unix time)
   */
  public to: number

  /**
   * The subscribers of the entry, gets notified on status changes
   *
   * @private
   */
  private subs: ((a: boolean|string) => void)[] = []

  /**
   * The status of the entry
   * Represents whether or not the underlying file is fully written to
   */
  public status: boolean|string = false

  /**
   * The path to the underlying file
   */
  public localPath?: string

  /**
   * The amount of current uses of the entry
   *
   * @private
   */
  private useCount: number = 1

  /**
   * StreamStateEntry constructor
   *
   * @param {string} vessel The vessel it represents
   * @param {string} variable The variable it represents
   * @param {number} from The start of the entrys interval, in unix time
   * @param {number} to The end of the entrys interval, in unix time
   */
  constructor (vessel: string, variable: string, from: number, to: number) {
    this.vessel = vessel
    this.variable = variable
    this.from = from
    this.to = to
  }

  /**
   * Updates the entrys filepath
   *
   * @param {string} path The absolute path of the file containing the stream
   */
  public setPath (path: string) {
    this.localPath = path
  }

  /**
   * Subtracts one from the entrys uses
   * Self-destructs once uses reach 0
   *
   * @param {StreamStateEntry} this streamstate entry
   */
  public subUse (this: StreamStateEntry) {
    this.useCount--

    if (this.useCount <= 0) {
      this.subs.forEach(sub => sub('Entry closing'))
      StreamStateController.remove(this)
    }
  }

  /**
   * Adds a use to the entrys uses
   */
  public addUse () {
    this.useCount++
  }

  /**
   * Updates the entrys status and notifies subscribers
   *
   * @param {boolean|string} status The new status of the underlying file
   */
  public setStatus (status: boolean|string) {
    if (status === this.status) return
    this.subs.forEach(sub => sub(status))
    this.status = status
  }

  /**
   * Add subscriber to the entry
   *
   * @param {Partial<SubFunc>} func A function to be called on status changes
   */
  public sub (func: (a: boolean|string) => void) {
    this.subs.push(func)
  }

  /**
   * Checks that the given entry is the same
   *
   * @param {StreamStateEntry} this This streamstate tnery
   * @param {StreamStateEntry} subject Another stream state entry
   * @returns {boolean} true if the entry matches
   */
  public equals (this: StreamStateEntry, subject: StreamStateEntry): boolean {
    if (this.from !== subject.from) {
      return false
    }
    if (this.to !== subject.to) {
      return false
    }
    if (this.variable !== subject.variable) {
      return false
    }
    if (this.vessel !== subject.vessel) {
      return false
    }
    if (this.useCount !== subject.useCount) {
      return false
    }
    if (this.localPath !== subject.localPath) {
      return false
    }
    return true
  }
}

/**
 * Class containing the logic to maintain the state of all streams
 */
export default class StreamStateController {
  private static streamState: {
        [key: string]: {
            [key: string]: {
                entries: StreamStateEntry[],
                addSubs: Record<string, SubFunc>,
                removeSubs: Record<string, SubFunc>
            }
        }
    } = {}

  /**
   * Returns all streamstate entries matching the parameters
   *
   * @param {string} vessel The vessel to look in
   * @param {string} variable The variable to look for
   * @param {number} from The start of the relevant range
   * @param {number} to The end of the relevant range
   * @returns {StreamStateEntry[]} All streamstate entries colliding with the given range
   */
  public static getInRange (vessel: string, variable: string, from: number, to: number): StreamStateEntry[] {
    return this.streamState[vessel]?.[variable]?.entries?.filter(entry => {
      return (from <= entry.from && entry.from < to) ||
                (from < entry.to && entry.to <= to)
    }).map(entry => {
      entry.addUse()
      return entry
    }) ?? []
  }

  /**
   * Add {@see SubFunc} to subscribers, called when entries are added
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {Partial<SubFunc>} func The sub func itself
   * @returns {SubFunc} The given function with an assigned id
   */
  public static subOnAdd (vessel: string, variable: string, func: Partial<SubFunc>): SubFunc {
    return this.sub(vessel, variable, func, 'add')
  }

  /**
   * Add {@see SubFunc} to subscribers, called when entries are removed
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {Partial<SubFunc>} func The sub func itself
   * @returns {SubFunc} The given function with an assigned id
   */
  public static subOnRemove (vessel: string, variable: string, func: Partial<SubFunc>): SubFunc {
    return this.sub(vessel, variable, func, 'remove')
  }

  /**
   * Utility function to add subfunc to subscribers
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {Partial<SubFunc>} func The sub func itself
   * @param {'add'|'remove'} type The type of action to sub to
   * @returns {SubFunc} The subfunc including an identifier
   */
  private static sub (vessel: string, variable: string, func: Partial<SubFunc>, type: 'add'|'remove'): SubFunc {
    if (this.streamState[vessel] == null) {
      this.streamState[vessel] = {}
    }
    if (this.streamState[vessel]?.[variable] == null) {
      this.streamState[vessel][variable] = { entries: [], addSubs: {}, removeSubs: {} }
    }
    const rands = faker.random.randomWord().substring(0, 2)
    func.id = Date.now() + '' + rands
    if (type === 'add') this.streamState[vessel][variable].addSubs[func.id] = func as SubFunc
    else this.streamState[vessel][variable].removeSubs[func.id] = func as SubFunc

    return func as SubFunc
  }

  /**
   * Remove subscriber from "add" subscription
   * Must be the returned subfunc from subscribtion with its unique id
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {SubFunc} func The sub func itself
   */
  public static unsubOnAdd (vessel: string, variable: string, func: SubFunc) {
    this.unsub(vessel, variable, func, 'add')
  }

  /**
   * Remove subscriber from "remove" subscription
   * Must be the returned subfunc from subscribtion with its unique id
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {SubFunc} func The sub func itself
   */
  public static unsubOnRemove (vessel: string, variable: string, func: SubFunc) {
    this.unsub(vessel, variable, func, 'remove')
  }

  /**
   *  Utility function to remove subfunc from subscribers
   *
   * @param {string} vessel The vessel to sub to
   * @param {string} variable The variable to sub to
   * @param {SubFunc} func The sub func itself
   * @param {'add'|'remove'} type What type of sub it is
   */
  private static unsub (vessel: string, variable: string, func: SubFunc, type: 'add'|'remove') {
    if (this.streamState[vessel]?.[variable] == null) return
    const mapping = type === 'add'
      ? this.streamState[vessel][variable].addSubs
      : this.streamState[vessel][variable].removeSubs
    delete mapping[func.id]
  }

  /**
   * Retrieves or creates an entry based on the parameters
   *
   * @param {string} vessel The vessel to look in
   * @param {string} variable The variable to look for
   * @param {number} from The start of the relevant range
   * @param {number} to The end of the relevant range
   * @returns {StreamStateEntry} The retrieved or newly created streamstate entry
   */
  public static getOrCreate (vessel: string, variable: string, from: number, to: number): StreamStateEntry {
    const gotten = this.get(vessel, variable, from, to)
    if (gotten != null) {
      gotten.addUse()
      return gotten
    }

    return this.create(vessel, variable, from, to)
  }

  /**
   * Creates entry for the given parameters
   *
   * @param {string} vessel The vessel to look in
   * @param {string} variable The variable to look for
   * @param {from} from The start of the relevant range
   * @param {number} to The end of the relevant range
   * @returns {StreamStateEntry} The newly created streamstate entry
   * @throws {Error} If created stream collides with existing stream(s)
   */
  public static create (vessel: string, variable: string, from: number, to: number): StreamStateEntry {
    const entry = new StreamStateEntry(vessel, variable, from, to)
    if (this.streamState[vessel] == null) {
      this.streamState[vessel] = { [variable]: { entries: [], addSubs: {}, removeSubs: {} } }
    } else if (this.streamState[vessel][variable] == null) {
      this.streamState[vessel][variable] = { entries: [], addSubs: {}, removeSubs: {} }
    }

    if (this.streamState[vessel][variable].entries.findIndex(entry => {
      return (from <= entry.from && to > entry.from) ||
                (from < entry.to && entry.to <= to)
    }) !== -1) {
      throw new Error('Streams in this interval already exist')
    }
    this.streamState[vessel][variable].entries.push(entry)
    for (let i = 0; i < Object.keys(this.streamState[vessel][variable].addSubs).length; i++) {
      const key = Object.keys(this.streamState[vessel][variable].addSubs)[i]
      const sub = this.streamState[vessel][variable].addSubs[key]
      entry.addUse()
      Promise.resolve(sub(entry)).then(() => entry.subUse())
    }
    return entry
  }

  /**
   * Retrieves an entry based on the parameters
   *
   * @param {string} vessel The vessel to look in
   * @param {string} variable The variable to look for
   * @param {number} from The start of the relevant range
   * @param {number} to The end of the relevant range
   * @returns {StreamStateEntry|undefined} A streamstate entry, if found
   */
  public static get (vessel: string, variable: string, from: number, to: number): StreamStateEntry|undefined {
    return this.streamState[vessel]?.[variable]?.entries?.find(entry => {
      return entry.from === from && entry.to === to
    })
  }

  /**
   * Removes the given entry from the state
   *
   * @param {StreamStateEntry} entry The entry to remove
   * @throws {Error} If the entry in its place does not match or does not exist
   */
  public static remove (entry: StreamStateEntry) {
    const eIdx = this.streamState[entry.vessel]?.[entry.variable]?.entries?.findIndex(rEntry => {
      return entry.from === rEntry.from && entry.to === rEntry.to
    })
    if (eIdx < 0 || eIdx == null) throw new Error('Entry does not exist')
    const stateEntry = this.streamState[entry.vessel][entry.variable]?.entries[eIdx]
    if (!entry.equals(stateEntry)) {
      throw new Error('Entries do not match')
    }
    this.streamState[entry.vessel][entry.variable].entries.splice(eIdx, 1)

    Object.keys(this.streamState[entry.vessel][entry.variable].removeSubs).forEach(sub => {
      this.streamState[entry.vessel][entry.variable].removeSubs[sub](entry)
    })

    if (fs.existsSync(entry.localPath!)) fs.unlinkSync(entry.localPath!)

    if (this.streamState[entry.vessel][entry.variable].entries.length === 0) {
      delete this.streamState[entry.vessel][entry.variable]
    }
    if (Object.keys(this.streamState[entry.vessel]).length === 0) {
      delete this.streamState[entry.vessel]
    }
  }
}
