import mqtt from 'mqtt'
import { clearTimeout } from 'timers'

/**
 * The default callback to run on connect
 */
const defaultCB = () => {
  console.log('MQTT connected\n')
}

/**
 * Creates an MQTT client and connects it to the broker
 *
 * @param {() => void} onConnectCB Optional callback to run on connect
 * @returns {Promise<mqtt.MqttClient>} A preconfigured mqtt client
 */
const getMQTT = async (onConnectCB: () => void = defaultCB): Promise<mqtt.MqttClient> => {
  return await new Promise<mqtt.MqttClient>((resolve, reject) => {
    try {
      const mqttClient = mqtt.connect('', {
        host: process.env.MQTT_BROKER_ADDRESS,
        protocol: process.env.MQTT_REQUIRE_TLS !== 'false' ? 'mqtts' : 'mqtt',
        username: process.env.MQTT_USER ?? undefined,
        password: process.env.MQTT_PWD ?? undefined,
        port: process.env.MQTT_PORT != null ? parseInt(process.env.MQTT_PORT) : process.env.MQTT_REQUIRE_TLS !== 'false' ? 8883 : 1883,
        connectTimeout: 1e3,
        reconnectPeriod: 1e3
      })

      const timeout = setTimeout(() => {
        mqttClient.emit('error', 'Cannot reach broker')
      }, 1e4)

      mqttClient.on('connect', async () => {
        clearTimeout(timeout)
        onConnectCB()
        resolve(mqttClient)
      })

      mqttClient.on('error', err => {
        reject(err)
        mqttClient.end()
      })
    } catch (e: any) {
      if (e instanceof Error) {
        reject(e)
      } else {
        reject(new Error(e))
      }
    }
  })
}
export default getMQTT
