import mqtt from 'mqtt'
import getMQTT from './getMQTTClient'

export interface MqttPoolClient extends mqtt.MqttClient {

}

/**
 * Generic FIFO queue datastructure
 *
 * @template T
 */
export class Queue<T> {
  /**
   * The number of nodes in the queue
   */
  public length = 0

  /**
   * The head node of the queue, removed on dequeue
   *
   * @private
   */
  private head: QueueNode<T>|undefined

  /**
   * The tail node of the queue, replaced on enqueue
   *
   * @private
   */
  private tail: QueueNode<T>|undefined

  /**
   * Enqueues an object by the queues type
   *
   * @param {T} value The value to enqueue
   */
  public enqueue (value: T): void {
    this.length++
    if (this.head == null) {
      this.head = new QueueNode<T>(value)
      this.tail = this.head
    } else {
      this.tail?.setNext(new QueueNode<T>(value))
      this.tail = this.tail?.getNext()
    }
  }

  /**
   * Dequeeus the head object of the queue
   *
   * @returns {T} The "first in" object in the queue
   * @throws {Error} If queue is empty
   */
  public dequeue (): T {
    if (this.length === 0) throw new Error('Queue is empty')
    this.length--
    const node = this.head!
    this.head = this.head?.getNext()

    if (this.length === 0) this.tail = undefined

    return node.getValue()
  }
}

/**
 * Class representing internal node structure of {@see Queue}
 *
 * @template T
 */
export class QueueNode<T> {
  /**
   * The node following this one
   *
   * @private
   */
  private next: QueueNode<T>|undefined

  /**
   * The value this node represents
   *
   * @private
   */
  private value: T

  /**
   * QueueNode constructor
   *
   * @param {T} value The value the node encompasses
   */
  constructor (value: T) {
    this.value = value
  }

  /**
   * Updates the nodes "next" value
   *
   * @param {QueueNode<T>} node The node to set as next
   */
  public setNext (node: QueueNode<T>) {
    this.next = node
  }

  /**
   * Returns the nodes next value
   *
   * @returns {QueueNode<T>|undefined} The next value
   */
  public getNext (): QueueNode<T>|undefined {
    return this.next
  }

  /**
   * Returns the value the node encompasses
   *
   * @returns {T} The value
   */
  public getValue (): T {
    return this.value
  }
}

/**
 * Class to pool mqtt clients
 */
export default class MqttPool {
  /**
   * The maximum number of clients at one time
   *
   * @private
   */
  private static readonly maxClients = 20

  /**
   * The current number of mqtt clients, idle or in use
   *
   * @private
   */
  private static clientCount = 0

  /**
   * FIFO queue of idle mqtt clients
   *
   * @private
   */
  private static queue: Queue<MqttPoolClient> = new Queue()

  /**
   * Resolve functions waiting for a client to become ready, FIFO
   *
   * @private
   */
  private static waiting: Queue<(client?: MqttPoolClient) => void> = new Queue()

  /**
   * Returns an MQTT client
   * If less than 20 mqtt clients exist, and none are available, a new client will be created
   *
   * @returns {MqttPoolClient} An MQTT client
   */
  public static async get (): Promise<MqttPoolClient> {
    return new Promise<MqttPoolClient>(async (resolve, reject) => {
      let client: MqttPoolClient|undefined

      while (client == null) {
        try {
          // First check if a client is on standby
          if (this.queue.length !== 0) {
            const dequeued = this.queue.dequeue()
            if (dequeued.disconnected || dequeued.disconnecting) {
              this.clientCount--
              continue
            } else {
              client = dequeued
            }

            // If not, check if adding clients is allowed
          } else if (this.clientCount < this.maxClients) {
            client = await getMQTT(() => {})
            this.clientCount++

            // Wait for a new client to be released
          } else {
            client = await new Promise<MqttPoolClient|undefined>(resolve => {
              this.waiting.enqueue(client => {
                resolve(client)
              })
            })
          }
        } catch (e) {
          return reject(e)
        }
      }
      resolve(client)
    })
  }

  /**
   * Returns mqtt client to the pool after clearing its event listeners.
   * If a faulty client is attempted released, it will be dropped and a new client will take its place
   *
   * @param {MqttPoolClient} client The client to release
   */
  public static release (client: MqttPoolClient) {
    // If the client is faulty, don't add it back to the queue, but subtract one from the clientcount
    if (client.disconnecting || client.disconnected || !client.connected) {
      this.clientCount--

      // If any processes are waiting for a client, let them create a new client by resolving undefined
      if (this.waiting.length !== 0) {
        this.waiting.dequeue()()
      }
      return
    }

    // Remove any listeners or subscriptions before adding back to the queue
    client.removeAllListeners('message')
    client.removeAllListeners('end')
    client.removeAllListeners('error')

    client.on('error', (e) => {
      console.log(e)
      client.end()
    })
    client.unsubscribe('#')

    if (this.waiting.length === 0) {
      this.queue.enqueue(client)
    } else {
      this.waiting.dequeue()(client)
    }
  }
}
