import getMQTT from '../services/mqtt/getMQTTClient'
import getMockMQTT from '../.jest/getMockMQTT'
import dataStreamer from '../services/database/dataStreamer'
import { MetaDataAggregator } from '../services/database/metaDataAggregator'
import faker from '@faker-js/faker'
import fetchRange from '../handlers/data/getRange/fetchRange'
import EntityDataController from '../services/database/entityDataController'
import path from 'path'
import StreamStateController from '../services/database/streamStateController'
import fs from 'fs'
import initModels from '../.jest/initModels'
import MqttPool, {MqttPoolClient} from '../services/mqtt/mqttPool'

jest.setTimeout(1e4)

jest.mock('../services/mqtt/getMQTTClient')
const mqttMock = getMQTT as jest.MockedFunction<typeof getMQTT>
mqttMock.mockImplementation(getMockMQTT)

jest.mock('../services/mqtt/mqttPool', () => {
  const actual = jest.requireActual('../services/mqtt/mqttPool')
  actual.default.release = (client: MqttPoolClient) => client.end(true)
  return actual
})

jest.mock('../services/database/dataStreamer')
const sdMock = dataStreamer as jest.MockedFunction<typeof dataStreamer>
sdMock.mockResolvedValue()

jest.mock('../services/database/metaDataAggregator')
const mdaMock = MetaDataAggregator as jest.Mocked<typeof MetaDataAggregator>
mdaMock.get.mockImplementation(() => {
  return {
    run: () => {},
    add: () => {},
    aggregate: () => {}
  } as unknown as MetaDataAggregator
})

const spyDate = jest.spyOn(Date, 'now')

beforeAll(async () => {
  await initModels()
})

it('Should fetch the data of given variable within the given range', async () => {
  const name = faker.name.lastName()
  const veNode = await EntityDataController.vesselByName(name)
  const vaNode = await EntityDataController.variableByName(name)
  const from = faker.datatype.number({ min: 0, max: 10000 })
  const to = faker.datatype.number({ min: 10000, max: 20000 })

  const res = await fetchRange(veNode, vaNode, { from, to })
  expect(res.name).toBe(vaNode.name)
  expect(Object.keys(res.double).length).not.toBe(0)

  // It should add a use to both nodes to avoid deletion before aggregation is complete
  expect(veNode.uses).toBe(2)
  expect(vaNode.uses).toBe(2)

  // It should create a streamstate entry
  const streamEntry = StreamStateController.get(veNode.name, vaNode.name, from, to)!
  expect(streamEntry).toBeDefined()
  expect(streamEntry.from).toBe(from)
  expect(streamEntry.to).toBe(to)
})

it('Should drop requests where from is larger than to', async () => {
  const name = faker.name.lastName()
  const veNode = await EntityDataController.vesselByName(name)
  const vaNode = await EntityDataController.variableByName(name)

  const res = await fetchRange(veNode, vaNode, { from: 100, to: 10 })
  expect(res.name).toBe(vaNode.name)
  expect(Object.keys(res.double).length).toBe(0)
})

it('Should continue searching if current time tmp file is taken', async () => {
  const name = faker.name.lastName()
  const veNode = await EntityDataController.vesselByName(name)
  const vaNode = await EntityDataController.variableByName(name)

  fs.writeFileSync(path.resolve(process.cwd(), './tmp/123.csv'), 'taken')
  spyDate.mockReturnValueOnce(123)

  await expect(fetchRange(veNode, vaNode, { from: 100, to: 1010 })).resolves

  spyDate.mockReset()
})
