import { testVariables, testVessel } from '../.jest/testConst'
import faker from '@faker-js/faker'
import getRange from '../handlers/data/getRange/getRange'
import getMQTT from '../services/mqtt/getMQTTClient'
import Reading, { ReadingBean } from '../models/reading'
import { MetaDataAggregator } from '../services/database/metaDataAggregator'
import initModels from '../.jest/initModels'
import pg from 'pg'
import byteCounter from '../services/logging/byteCounter'
import getMockMQTT from '../.jest/getMockMQTT'
import { CreationAttributes, InferCreationAttributes } from 'sequelize'
import dataStreamer from '../services/database/dataStreamer'
import MetaData from '../models/metaData'
import EntityDataController from '../services/database/entityDataController'
import MqttPool, { MqttPoolClient } from '../services/mqtt/mqttPool'

jest.setTimeout(1e4)

jest.mock('../services/mqtt/getMQTTClient')
const mqttMock = getMQTT as jest.MockedFunction<typeof getMQTT>
mqttMock.mockImplementation(getMockMQTT)

jest.mock('../services/mqtt/mqttPool', () => {
  const actual = jest.requireActual('../services/mqtt/mqttPool')
  actual.default.release = (client: MqttPoolClient) => client.end(true)
  return actual
})

jest.mock('../services/database/metaDataAggregator')
const mdaMock = MetaDataAggregator as jest.Mocked<typeof MetaDataAggregator>
mdaMock.get.mockImplementation(() => {
  return {
    /**
     *
     */
    run: () => {},
    /**
     *
     */
    add: () => {},
    /**
     *
     */
    aggregate: () => {}
  } as unknown as MetaDataAggregator
})

jest.mock('../services/database/dataStreamer')
const sdMock = dataStreamer as jest.MockedFunction<typeof dataStreamer>
sdMock.mockResolvedValue()

const actualLoadRange = jest.requireActual('../services/database/loadRange')
const lrMock = jest.spyOn(actualLoadRange, 'default')

beforeAll(async () => {
  pg.defaults.parseInt8 = true
  await initModels()
})

let range: {
    from: number,
    to: number
}|undefined
const names = testVariables
const vessel = testVessel
const recordCount = 1000

beforeEach(async () => {
  range = {
    from: Date.now() - 1e3,
    to: Date.now() + 1e3
  }
  const vesselNode = await EntityDataController.vesselByName(vessel)
  const rangeInterval = (range.to - range.from) / recordCount
  const allRecords: ReadingBean[] = []
  const metadata: InferCreationAttributes<MetaData, {omit: 'id'}>[] = []
  for (let i1 = 0; i1 < names.length; i1++) {
    const name = names[i1]
    const node = await EntityDataController.variableByName(name)
    const typeRecords: ReadingBean[] = []

    for (let i = 0; i < recordCount; i++) {
      typeRecords.push({
        vessel: vesselNode.id,
        variable: node.id,
        value: faker.datatype.number(),
        timestamp: range.from + (i + 1) * rangeInterval
      })
    }
    allRecords.push(...typeRecords)
    metadata.push({
      vessel: vesselNode.id,
      variable: node.id,
      from: range!.from + 2e2,
      to: range!.to - 2e2
    })
  }
  try {
    await Reading.bulkCreate(allRecords as CreationAttributes<Reading>[], { logging: false, ignoreDuplicates: true })

    await MetaData.bulkCreate(metadata, { logging: false })
  } catch (e) {
    console.log(e)
  }
})

it('Should get all cached data', async () => {
  const dataCost = byteCounter()
  const records = await getRange(vessel, names, range!, 10, dataCost)
  expect(dataCost.get()).not.toBe(0)
  expect(records.length).toBe(names.length)
  records.forEach(signal => {
    expect(names.includes(signal.name)).toBe(true)
    expect(Object.keys(signal.double).length).toBeGreaterThan(0)
    Object.keys(signal.double).map(k => parseInt(k)).forEach(k => {
      expect(k).toBeGreaterThanOrEqual(range!.from)
      expect(k).toBeLessThanOrEqual(range!.to)
    })
  })

  expect(lrMock).toHaveBeenCalled()

  const fromCache = records.findIndex(record => {
    return Object.keys(record.double).map(d => parseInt(d)).findIndex(ts => {
      return ts > range!.from + 2e2 && ts < range!.to - 2e2
    }) !== -1
  }) !== -1
  expect(fromCache).toBeTruthy()
})
