import StreamStateController, { StreamStateEntry, SubFunc } from '../services/database/streamStateController'
import faker from '@faker-js/faker'

const vessel = faker.name.lastName()
const variable = faker.lorem.word()
let streams: StreamStateEntry[] = []

afterEach(() => {
  streams.forEach(stream => stream.subUse())
  streams = []
})

it('Should fetch the correct interval', () => {
  const stream1 = StreamStateController.create(vessel, variable, 0, 125)
  const stream2 = StreamStateController.create(vessel, variable, 175, 200)
  const stream3 = StreamStateController.create(vessel, variable, 200, 300)
  streams.push(stream1, stream2, stream3)

  const stream = StreamStateController.get(vessel, variable, 175, 200)
  expect(stream?.equals(stream2)).toBeTruthy()
})

it('Should get all streams touching the requested interval', () => {
  const stream1 = StreamStateController.create(vessel, variable, 0, 100)
  const stream2 = StreamStateController.create(vessel, variable, 100, 150)
  const stream3 = StreamStateController.create(vessel, variable, 175, 200)
  const stream4 = StreamStateController.create(vessel, variable, 200, 300)
  streams.push(stream1, stream2, stream3, stream4)

  const result = StreamStateController.getInRange(vessel, variable, 100, 200)
  expect(result.length).toBe(2)
  expect(result[0].equals(stream2)).toBeTruthy()
  expect(result[1].equals(stream3)).toBeTruthy()
})

it('Should not be possible to get an entry after its uses reach 0', () => {
  const stream = StreamStateController.create(vessel, variable, 0, 100)
  stream.subUse()

  expect(StreamStateController.get(vessel, variable, 0, 100)).toBeUndefined()
})

it('Should notify subscribers on status changes and on close', () => {
  const stream = StreamStateController.create(vessel, variable, 0, 100)
  let statusSub = false
  let closeSub = false
  stream.sub((a) => {
    if (typeof a === 'boolean') {
      statusSub = a
    } else {
      closeSub = true
    }
  })

  expect(statusSub).toBeFalsy()
  stream.setStatus(true)
  expect(statusSub).toBeTruthy()
  stream.setStatus(false)
  expect(statusSub).toBeFalsy()

  expect(closeSub).toBeFalsy()
  stream.subUse()
  expect(closeSub).toBeTruthy()
})

it('Should add a use if already exists with getOrCreate', () => {
  const stream = StreamStateController.create(vessel, variable, 0, 100)
  StreamStateController.getOrCreate(vessel, variable, 0, 100)

  let complete = false
  stream.sub(() => {
    complete = true
  })
  stream.subUse()
  expect(complete).toBeFalsy()
  stream.subUse()
  expect(complete).toBeTruthy()
})

it('Should notify when entry is added on add subs', () => {
  let addedNotified = false
  StreamStateController.subOnAdd(vessel, variable, function () {
    addedNotified = true
  } as Partial<SubFunc>)

  streams.push(StreamStateController.create(vessel, variable, 0, 100))

  expect(addedNotified).toBeTruthy()
})

it('Should remove the correct entry', () => {
  let sub1Called = false
  StreamStateController.subOnAdd(vessel, variable, function () {
    sub1Called = true
  } as Partial<SubFunc>)

  let sub2Called = false
  const sub2 = StreamStateController.subOnAdd(vessel, variable, function () {
    sub2Called = true
  } as Partial<SubFunc>)

  StreamStateController.unsubOnAdd(vessel, variable, sub2)
  streams.push(StreamStateController.create(vessel, variable, 0, 100))

  expect(sub1Called).toBeTruthy()
  expect(sub2Called).toBeFalsy()
})
