import pg from 'pg'
import initModels from '../.jest/initModels'
import wipeCache from '../handlers/commands/wipeCache'
import { Command } from '../proto/wlCommand'
import EntityDataController, { EntityData } from '../services/database/entityDataController'
import faker from '@faker-js/faker'
import Reading from '../models/reading'
import { Op, InferCreationAttributes } from 'sequelize'
import MetaData from '../models/metaData'

let nodes: EntityData[] = []

beforeAll(async () => {
  pg.defaults.parseInt8 = true
  await initModels()
})

afterEach(async () => {
  await Reading.destroy({
    where: {
      variable: { [Op.in]: nodes.filter(node => node.type === 'variable').map(node => node.id) },
      vessel: { [Op.in]: nodes.filter(node => node.type === 'vessel').map(node => node.id) }
    },
    logging: false
  })
  nodes = []
})

it('Should wipe the entire database', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.name.lastName())
  nodes.push(vesselNode, variableNode)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: faker.datatype.number({ min: 0, max: 1e5, precision: 1 }),
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  // Add data and ensure it is in the database
  await Reading.bulkCreate(readings, { logging: false })
  await expect(Reading.findAll({ where: { variable: variableNode.id, vessel: vesselNode.id }, logging: false }).then(result => result.length)).resolves.toBeGreaterThan(0)

  // Wipe cache, then make sure the data is now gone
  const command: Command = {
    action: 'wipe',
    variables: []
  }
  await wipeCache(command)
  await expect(Reading.findAll({ where: { variable: variableNode.id, vessel: vesselNode.id }, logging: false }).then(result => result.length)).resolves.toBe(0)
})

it('Should wipe data of the given variable names', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.name.lastName())
  const variableNode2 = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode, variableNode2)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    }, {
      vessel: vesselNode.id,
      variable: variableNode2.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  try {
    // Add data and ensure it is in the database
    await Reading.bulkCreate(readings, { logging: false })
    await expect(Reading.findAll({
      where: { variable: { [Op.in]: [variableNode.id, variableNode2.id] } },
      logging: false
    }).then(result => result.length)).resolves.toBe(100)
  } catch (e) {
    console.log(e)
  }
  // Wipe cache, then make sure the data is now gone
  const command: Command = {
    action: 'wipe',
    variables: [variableNode2.name]
  }
  await wipeCache(command)
  const after = await Reading.findAll({ where: { variable: { [Op.in]: [variableNode.id, variableNode2.id] } }, logging: false })
  expect(after.length).toBe(50)
  expect(after.filter(reading => reading.variable === variableNode2.id).length).toBe(0)
})

it('Should wipe data based on the given range', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  // Add data and ensure it is in the database
  await Reading.bulkCreate(readings, { logging: false })
  await expect(Reading.findAll({ where: { variable: variableNode.id, vessel: vesselNode.id }, logging: false }).then(result => result.length)).resolves.toBeGreaterThan(0)

  // Wipe cache, then make sure the data is now gone
  const command: Command = {
    action: 'wipe',
    variables: [],
    ranges: {
      from: 50 + '',
      to: 100 + ''
    }
  }
  await wipeCache(command)
  await expect(Reading.findAll({ where: { variable: variableNode.id, vessel: vesselNode.id }, logging: false }).then(result => result.length)).resolves.toBe(44)
})

it('Should delete readings based on the given vessel', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const vesselNode2 = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode, vesselNode2)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    }, {
      vessel: vesselNode2.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  // Add data and ensure it is in the database
  await Reading.bulkCreate(readings, { logging: false })
  await expect((await Reading.findAll({ where: { vessel: { [Op.in]: [vesselNode.id, vesselNode2.id] } }, logging: false })).length).toBe(100)

  // Wipe cache, then make sure the data is now gone
  const command: Command = {
    action: 'wipe',
    variables: [],
    vessel: vesselNode2.name
  }
  await wipeCache(command)
  await expect((await Reading.findAll({ where: { vessel: { [Op.in]: [vesselNode.id, vesselNode2.id] } }, logging: false })).length).toBe(50)
})

it('Should delete before and after the given from and to in separate calls', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  // Add data and ensure it is in the database
  await Reading.bulkCreate(readings, { logging: false })
  await expect(Reading.findAll({
    where: {
      vessel: vesselNode.id,
      variable: variableNode.id
    },
    logging: false
  }).then(result => result.length)).resolves.toBeGreaterThan(0)

  // Wipe cache after from
  const command1: Command = {
    action: 'wipe',
    variables: [],
    ranges: {
      from: 450 + '',
      to: undefined
    }
  }
  await wipeCache(command1)
  await expect(Reading.findAll({
    where: {
      vessel: vesselNode.id,
      variable: variableNode.id
    },
    logging: false
  }).then(result => result.length)).resolves.toBe(45)

  // Wipe cache before to
  const command2: Command = {
    action: 'wipe',
    variables: [],
    ranges: {
      from: undefined,
      to: 40 + ''
    }
  }
  await wipeCache(command2)
  await expect(Reading.findAll({
    where: {
      vessel: vesselNode.id,
      variable: variableNode.id
    },
    logging: false
  }).then(result => result.length)).resolves.toBe(40)
})

it('Should wipe only relevant data with all parameters', async () => {
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const vesselNode2 = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.lorem.word())
  const variableNode2 = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode, vesselNode2, variableNode2)

  const readings: InferCreationAttributes<Reading>[] = []
  for (let i = 0; i < 50; i++) {
    readings.push({
      vessel: vesselNode.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    }, {
      vessel: vesselNode2.id,
      variable: variableNode.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    }, {
      vessel: vesselNode.id,
      variable: variableNode2.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    }, {
      vessel: vesselNode2.id,
      variable: variableNode2.id,
      timestamp: i * 10,
      value: faker.datatype.number({ precision: 0.001 })
    })
  }

  // Add data and ensure it is in the database
  await Reading.bulkCreate(readings, { logging: false })
  await expect((await Reading.findAll({ where: { vessel: { [Op.in]: [vesselNode.id, vesselNode2.id] } }, logging: false })).length).toBe(200)

  // Wipe cache, then make sure the data is now gone
  const command: Command = {
    action: 'wipe',
    variables: [variableNode.name],
    vessel: vesselNode2.name,
    ranges: {
      from: 0 + '',
      to: 40 + ''
    }
  }
  await wipeCache(command)
  await expect((await Reading.findAll({ where: { vessel: { [Op.in]: [vesselNode.id, vesselNode2.id] } }, logging: false })).length).toBe(195)
})

it('Should handle metadata', async () => {
  // Vessel/variable changeups have been tested in the preceding tests, so we only need to test with ranges in mind
  // since both metadata and reading queries use the same values for them
  const vesselNode = await EntityDataController.vesselByName(faker.name.lastName())
  const variableNode = await EntityDataController.variableByName(faker.lorem.word())
  nodes.push(vesselNode, variableNode)

  /*
  Add some intervals, the fact they make an invalid state doesn't matter for testing
   */
  await MetaData.bulkCreate([{
    vessel: vesselNode.id,
    variable: variableNode.id,
    from: 0,
    to: 101
  }, {
    vessel: vesselNode.id,
    variable: variableNode.id,
    from: 100,
    to: 200
  }, {
    vessel: vesselNode.id,
    variable: variableNode.id,
    from: 199,
    to: 300
  }, {
    vessel: vesselNode.id,
    variable: variableNode.id,
    from: 0,
    to: 300
  }], {logging: false})
  // Wipe cache, then make sure the metadata is now gone
  const command: Command = {
    action: 'wipe',
    variables: [variableNode.name],
    vessel: vesselNode.name,
    ranges: {
      from: 100 + '',
      to: 200 + ''
    }
  }
  await wipeCache(command)

  const all = await MetaData.findAll({ where: { vessel: vesselNode.id, variable: variableNode.id }, logging: false })
  const pre = all.filter(m => m.from === 0 && m.to === 100)
  const post = all.filter(m => m.from === 200 && m.to === 300)
  // Should include first 0-100 array and 0-100 from split 0-300
  expect(pre.length).toBe(2)

  // Should include first 200-300 array and 200-300 from split 0-300
  expect(post.length).toBe(2)
})
