import { Box, Flex, useColorMode, Spinner, useColorModeValue } from '@chakra-ui/react'
import type { ChartData, TickOptions } from 'chart.js'
import { Chart as ChartJS, LinearScale, TimeScale, TimeSeriesScale, LineController, PointElement, LineElement, Legend, Tooltip, Title, CategoryScale } from 'chart.js'
import { ReactElement, useEffect, useState } from 'react'
import { Chart } from 'react-chartjs-2'
import zoom from 'chartjs-plugin-zoom'
import { base, options, ticks } from './options'
import 'chartjs-adapter-date-fns'
import { useSignalContext } from '../signal/context'
import ChartNavigation from './navigation'
import SessionUsage from './sessionusage'
import { Signal } from '../../globals'

// Register all ChartJS components to the window module
ChartJS.register(zoom, LinearScale, TimeScale, TimeSeriesScale, LineController, PointElement, LineElement, Legend, Tooltip, Title, CategoryScale)

interface Props {
  handleFetch: Function
}

/**
 * TimeChart Functional Component
 *
 * @param {Props} args TimeChart component props
 * @returns {ReactElement<Props>} react element
 */
const TimeChart = ({ handleFetch }: Props): ReactElement<Props> => {
  const { colorMode } = useColorMode()
  const { signals, isLoading, chartRef, expanded } = useSignalContext()
  const [chartData, setChartData] = useState<ChartData<'bar'> | null>(null)
  const [chartOptions, setChartOptions] = useState<any>({ ...base })

  /**
   * Helper method for getting string array of all signal names
   *
   * @returns {string[]} signal names
   * @internal
   */
  const getSignalNames = (): string[] => {
    if (!signals) return []
    return signals.map((signal: Signal) => signal.name)
  }

  /**
   * Method for getting wanted color based on dark mode
   *
   * @param {boolean} dark dark mode variable
   * @returns {string} rgba string
   * @internal
   */
  const color = (dark: boolean): string => (dark ? 'rgba(0, 0, 0, 0.18)' : 'rgba(0, 0, 0, 0.12)')

  /**
   * Method for generating a tick option object for color based on dark mode
   *
   * @param {boolean} dark dark mode variable
   * @returns {Partial<TickOptions>} partial tick option object
   */
  const grid = (dark: boolean): { color: string; borderColor: string } => ({
    color: color(dark),
    borderColor: color(dark)
  })

  useEffect(() => {
    const chart = chartRef.current
    if (chart) {
      const dark: boolean = colorMode === 'dark'

      // Change tick color of x-scale
      if (chart.options.scales.x) {
        chart.options.scales.x.grid = grid(dark)
        chart.options.scales.x.ticks = {
          ...ticks,
          color: dark ? '#D0D3D4' : '#4D5656'
        }
      }

      // Change tick color of y-scales
      getSignalNames().forEach((name: string, idx: number) => {
        if (chart.options.scales[name]) {
          chart.options.scales[name].grid = {
            ...grid(dark),
            display: idx === 0
          }
        }
      })

      chart.stop()
      chart.update('none')
    }
  }, [colorMode])

  // UseEffect for initializing chart instance dataset format
  useEffect((): any => {
    const data: any = {
      datasets: []
    }
    setChartData(data)
  }, [])

  // UseEffect for coloring y-scales based on signals
  useEffect((): any => {
    const options_ = options(getSignalNames(), handleFetch)
    setChartOptions(options_)
  }, [signals])

  return (
    <Flex flexDir="column" mx={expanded ? 5 : 0} position="relative">
      <Box display={{ base: 'none', md: 'block' }}>
      <SessionUsage />
      </Box>
      <ChartNavigation handleFetch={handleFetch} />
      <Box h={{ base: expanded ? 'calc(100vh - 17rem)' : '35rem', md: expanded ? 'calc(100vh - 15rem)' : '35rem' }} w={expanded ? '100%' : { base: '90vw', md: '70vw', xl: '4xl' }} pos="relative">
        <Chart ref={chartRef} type="line" options={chartOptions} data={chartData || { datasets: [] }} />
        {isLoading && <Spinner pos={'absolute'} bottom={12} right={5} thickness="4px" speed="0.55s" emptyColor={'transparent'} color={useColorModeValue('gray.600', 'gray.400')} size="lg" />}
      </Box>
    </Flex>
  )
}

export default TimeChart
