import { IconButton, Flex } from '@chakra-ui/react'
import { ReactElement, useEffect, useRef } from 'react'
import { HiArrowSmDown, HiArrowSmLeft, HiArrowSmRight, HiArrowSmUp, HiRefresh, HiZoomIn, HiZoomOut } from 'react-icons/hi'
import { useSignalContext } from '../signal/context'

interface Props {
  handleFetch: Function
}

/**
 * Navigational controls for chart component
 *
 * @param {Props} args arguments for functional component
 * @returns {ReactElement<Props>} functional react node
 */
const ChartNavigation = ({ handleFetch }: Props): ReactElement<Props> => {
  const { chartRef } = useSignalContext()
  const refLeft = useRef<HTMLButtonElement>(null)
  const refRight = useRef<HTMLButtonElement>(null)
  const refUp = useRef<HTMLButtonElement>(null)
  const refDown = useRef<HTMLButtonElement>(null)

  /**
   * Method for focusing and clicking a given Button Element
   *
   * @param {HTMLButtonElement} button button ref
   */
  const handleFocusAndClick = (button: HTMLButtonElement) => {
    if (button) {
      button.focus()
      button.click()
    }
  }

  useEffect(() => {
    document.addEventListener(
      'keyup',
      (event: any) => {
        event.preventDefault()
        if (refLeft && refRight && refUp && refDown) {
          switch (event.keyCode) {
            case 37:
              handleFocusAndClick(refLeft.current)
              break
            case 38:
              handleFocusAndClick(refUp.current)
              break
            case 39:
              handleFocusAndClick(refRight.current)
              break
            case 40:
              handleFocusAndClick(refDown.current)
              break
          }
        }
      },
      false
    )
  }, [])

  /**
   * Method for changing zoom level of chart instance by given value
   *
   * @param {number} value mutator number
   */
  const handleZoom = (value: number) => {
    if (chartRef.current) {
      chartRef.current.zoom({ x: value })
    }
  }

  /**
   * Method for panning to x or y directions with given value
   *
   * @param {any} args mutator arguments
   */
  const handlePan = (args: any) => {
    if (chartRef.current) {
      chartRef.current.pan({ ...args }, undefined, 'normal')
    }
  }

  /**
   * Method for resetting zoom of chart component
   */
  const handleRestZoom = () => {
    if (chartRef.current) {
      chartRef.current.resetZoom()
    }
  }
  return (
    <Flex flexDir="row" gridGap={2}>
      <IconButton aria-label="refresh" onClick={handleRestZoom} icon={<HiRefresh />} />
      <IconButton aria-label="zoom-in" onClick={() => handleZoom(1.1)} icon={<HiZoomIn />} />
      <IconButton aria-label="zoom-out" onClick={() => handleZoom(0.9)} icon={<HiZoomOut />} />
      <IconButton ref={refLeft} aria-label="pan-left" onClick={() => handlePan({ x: 100 })} icon={<HiArrowSmLeft />} />
      <IconButton ref={refRight} aria-label="pan-right" onClick={() => handlePan({ x: -100 })} icon={<HiArrowSmRight />} />
      <IconButton ref={refUp} aria-label="pan-right" onClick={() => handlePan({ y: 100 })} icon={<HiArrowSmUp />} />
      <IconButton ref={refDown} aria-label="pan-right" onClick={() => handlePan({ y: -100 })} icon={<HiArrowSmDown />} />
    </Flex>
  )
}

export default ChartNavigation
