// @ts-ignore
import { en } from 'date-fns/locale'
import { colors } from '../../utils/constants'
import { ChartOptions, PluginChartOptions } from 'chart.js'
import { debounce } from '../../utils/tools'

// Memory for last min max limit when calling handleFetch
let limit = { min: 0, max: 0 }

// We have to debounce locally for panning
const fetch = debounce((data: any, handleFetch: Function) => {
  if (data.chart) handleFetch(limit, data.chart)
}, 500)

/**
 * Method for generating zoom and pan options for chartjs-zoom-plugin
 *
 * @param {Function} handleFetch function for fetching new data
 * @returns {Partial<PluginChartOptions>} zoom
 */
const zoom: any = (handleFetch: Function) => ({
  ...base.plugins.zoom,
  pan: {
    overScaleMode: 'y',
    enabled: true,
    mode: 'xy',
    speed: 10,
    /**
     * Handler for calling when chart is currently panning, this is for navigational buttons and arrow keys
     *
     * @param {any} data ChartJS ref
     * @returns {void}
     */
    onPan: (data: any) => fetch(data, handleFetch),
    /**
     * Handler for calling when a pan event has started
     *
     * @param {any} data ChartJS ref
     */
    onPanStart: (data: any) => {
      const { min, max } = data.chart.scales.x
      limit = { min, max }
    }
  },
  zoom: {
    overScaleMode: 'y',
    wheel: {
      enabled: true
    },
    pinch: {
      enabled: true
    },
    mode: 'xy',
    /**
     * Handler for calling when a zoom event has started
     *
     * @param {any} data ChartJS ref
     */
    onZoomStart: (data: any) => {
      const { min, max } = data.chart.scales.x
      limit = { min, max }
    },
    /**
     * Handler for calling when a zoom event has completed
     *
     * @param {any} data ChartJS ref
     */
    onZoomComplete: (data: any) => {
      if (data.chart) handleFetch(limit, data.chart)
    }
  }
})

const tooltip: any = {
  callbacks: {
    /**
     * Title callback function
     *
     * @param {any} tooltipItems all available items
     * @returns {any} title object
     */
    title: (tooltipItems: any) => {
      const item = tooltipItems[0]
      return item.label
    },
    /**
     * Label callback function
     *
     * @param {any} data ChartJS ref
     * @returns {any} label object
     */
    label: (data: any) => {
      return `${data.dataset.label}: ${data.formattedValue}`
    }
  }
}

/**
 * Method for generating individual scales for each signal
 *
 * @param {string[]} signals signal names
 * @returns {Partial<ChartOptions>} scales
 */
const scales: any = (signals: any[]): ChartOptions => {
  let y: { [key: string]: any } = {}
  signals.forEach((signal: string, idx) => {
    y = {
      ...y,
      [signal]: {
        grid: {
          display: idx === 0
        },
        position: 'left',
        ticks: {
          /**
           * Ticks callback function
           *
           * @param {any} val tick value
           * @param {number} index tick index
           * @returns {string} new tick value
           */
          callback: (val: any, index: number) => {
            // ES2020 added support for NumberFormatter, use this instead of importing npm package millify
            const formatter = Intl.NumberFormat('en', { notation: 'compact' })
            return formatter.format(val)
          },
          autoSkip: true,
          color: colors[idx]
        }
      }
    }
  })
  return { ...y }
}

export const base = {
  parsing: false,
  plugins: {
    zoom: {
      pan: {
        overScaleMode: 'y',
        enabled: true,
        mode: 'xy',
        speed: 10
      },
      zoom: {
        overScaleMode: 'y',
        wheel: {
          enabled: true
        },
        pinch: {
          enabled: true
        },
        mode: 'xy'
      }
    }
  }
}

export const ticks = {
  sampleSize: 40,
  maxTicksLimit: 10,
  offset: true,
  major: {
    enabled: false
  },
  autoSkip: true,
  maxRotation: 0
}

/**
 * Method for generating a options object for ChartJS instance
 *
 * @param {string[]} signals signalnames
 * @param {Function} handleFetch function for calling on new data
 * @returns {ChartOptions} chart options
 */
export const options: any = (signals: any[], handleFetch: Function) => ({
  responsive: true,
  maintainAspectRatio: false,
  parsing: false,
  interaction: {
    mode: 'nearest',
    intersect: false,
    axis: 'xy'
  },
  scales: {
    ...scales(signals),
    x: {
      type: 'time',
      time: {
        tooltipFormat: 'MMM dd HH:mm ss.SSSS',
        displayFormats: {
          millisecond: 'ss.SSSS',
          second: 'HH:mm:ss',
          minute: 'MMM dd HH:mm',
          hour: 'MMM dd HH:mm',
          day: 'MMM dd HH:mm',
          week: 'MMM dd HH:mm',
          month: 'MMM dd HH:mm',
          quarter: 'MMM dd HH:mm',
          year: 'HMMM dd HH:mm'
        }
      },
      adapters: {
        date: {
          locale: en
        }
      },
      ticks
    }
  },
  plugins: {
    tooltip,
    zoom: zoom(handleFetch)
  }
})
