import { Grid } from '@chakra-ui/react'
import { ReactElement } from 'react'
import { CommandMemory } from '../../../pages/commands'
import GridInput from '../grid/gridInput'

interface Props {
  memory: CommandMemory
  setMemory: any
}

/**
 * Functional Component for rendering Unix Panel component, this handles vessel, signals and range,
 * by letting user directly input 2 unix timestamps as from - to
 *
 * @param {Props} args Unix Panel props
 * @returns {ReactElement<Props>} functional component
 */
const UnixPanel = ({ memory, setMemory }: Props): ReactElement<Props> => {
  /**
   * Helper method for turning a Date variable into a unix timestamp
   * is based on given key, so that either "from" or "to" can be converted from a DateRange
   *
   * @param {CommandMemory} _memory current memory
   * @param {"from" | "to"} key accessor indicator
   * @returns {string} result
   */
  const getUnix = (_memory: CommandMemory, key: 'from' | 'to'): string => {
    try {
      return _memory.range[key].getTime().toString()
    } catch (err: any) {
      return ''
    }
  }

  /**
   * Helper Method for handling conversion between a unix timestamp and a Date object
   *
   * @param {string} value given timestamp
   * @returns {Date} converted timestamp
   */
  const handleValue = (value: string): Date => {
    let result = parseInt(value)
    if (isNaN(result)) result = 0
    return new Date(result)
  }

  /**
   * Helper Method for setting the memory when input has changed,
   * Utilizes helper methods for conversion
   *
   * @param {"from" | "to"} key accessor indicator
   * @returns {Function} callback for inputs onChange
   */
  const setUnix = (key: 'from' | 'to'): Function => {
    return (value: string) => {
      setMemory({
        ...memory,
        range: {
          ...memory.range,
          [key]: handleValue(value)
        }
      })
    }
  }

  return (
    <Grid gap="3" templateColumns="repeat(2, 1fr)">
      <GridInput colSpan={2} title="Vessel Name" accessor="vesselid" value={memory} setValue={setMemory} />
      <GridInput colSpan={2} title="Signals" accessor="signals" value={memory} setValue={setMemory} />
      <GridInput colSpan={1} title="From (UNIX)" value={getUnix(memory, 'from')} setValue={setUnix('from')} />
      <GridInput colSpan={1} title="To (UNIX)" value={getUnix(memory, 'to')} setValue={setUnix('to')} />
    </Grid>
  )
}

export default UnixPanel
