import { Box, Flex, Text, Button, useColorModeValue, Stack, useColorMode, Heading, HStack, useDisclosure, IconButton, Divider } from '@chakra-ui/react'
import { AddIcon, CloseIcon, HamburgerIcon, MoonIcon, SunIcon } from '@chakra-ui/icons'
import packageJson from '../../package.json'
import NavLink from '../shared/navlink'
import MobileNav from './mobilenav'
import { ReactElement, useEffect } from 'react'
import { navigation } from '../../utils/constants'
import { normalizeName } from '../../utils/tools'
import { useRouter } from 'next/router'
import { useSiteState } from '../../utils/context/state'
import NavigationHandler from '../../utils/handlers/navigationHandler'

/**
 * Functional Component for rendering NavBar
 *
 * @returns {ReactElement} functional component
 */
const NavBar = (): ReactElement => {
  const { colorMode, toggleColorMode } = useColorMode()
  const { isOpen, onOpen, onClose } = useDisclosure()
  const { vessels, set } = useSiteState()
  const router = useRouter()
  const { isActive, getPathPrefix, handleRemoveVessel } = NavigationHandler()

  useEffect(() => {
    // set scrollbar in chrome / safari to dark-theme
    document.documentElement.setAttribute('data-color-scheme', colorMode)
    return () => {}
  }, [colorMode])

  /**
   * Method for generating side items if vessels are found
   *
   * @returns {ReactElement | void} component
   */
  const buildSideItems = (): ReactElement | void => {
    if (!vessels) return null
    if (vessels.length === 0) return null
    return (
      <>
        {vessels.map((vessel: string, index: number) => (
          <Flex align="center">
            <NavLink key={index} active={isActive(vessel)} href={`${getPathPrefix(router.asPath)}/${vessel}/page/1`}>
              {vessel || 'Loading...'}
            </NavLink>
            <CloseIcon _hover={{ transform: 'scale(1.5)' }} ml={3} mr={2.5} cursor="pointer" transition="ease .2s" onClick={() => handleRemoveVessel(vessel)} w={3} h={3} />
          </Flex>
        ))}
      </>
    )
  }

  return (
    <>
      <Box bgGradient={useColorModeValue('linear(to-l, gray.50, gray.200)', 'linear(to-l, gray.800, gray.900)')} shadow="sm" zIndex={10} px={4} sx={{ position: 'sticky', top: '0' }}>
        <Flex mx="auto" maxW={'6xl'} h={16} alignItems={'center'} justifyContent={'space-between'}>
          <IconButton size={'md'} icon={isOpen ? <CloseIcon /> : <HamburgerIcon />} aria-label={'Open Menu'} display={{ md: 'none' }} onClick={isOpen ? onClose : onOpen} />
          <HStack spacing={8} alignItems={'center'}>
            <Box>
              <Heading size="md">{normalizeName(packageJson.name)}</Heading>
              <Text fontWeight="semibold" fontSize={14} opacity="0.7">{`v${packageJson.version}`}</Text>
            </Box>
            <HStack as={'nav'} spacing={4} display={{ base: 'none', md: 'flex' }}>
              {navigation.map((nav: any) => (
                <NavLink key={nav.title} href={nav.path}>
                  {nav.title}
                </NavLink>
              ))}
            </HStack>
          </HStack>
          <Flex alignItems={'center'} mr="16px">
            <Stack direction={'row'} spacing={7}>
              <Button onClick={toggleColorMode}>{colorMode === 'light' ? <MoonIcon /> : <SunIcon />}</Button>
            </Stack>
          </Flex>
        </Flex>
      </Box>
      <MobileNav isOpen={isOpen}>
        {navigation.map((nav: any) => (
          <NavLink key={nav.title} href={nav.path} active={false}>
            {nav.title}
          </NavLink>
        ))}
        <Divider borderColor={useColorModeValue('gray.400', 'gray.600')} />
        {buildSideItems()}
        <Box
          mt={5}
          px={2}
          py={1}
          cursor="pointer"
          outline="0 0 0 3px var(--chakra-colors-gray-500)"
          rounded={'md'}
          opacity={0.8}
          fontWeight={600}
          bg={useColorModeValue('gray.200', 'gray.700')}
          _hover={{
            bg: useColorModeValue('gray.300', 'gray.600')
          }}
          display="flex"
          alignItems="center"
          justifyContent="space-between"
          onClick={set.onOpen}
        >
          Add
          <AddIcon color="green.300" w={3.5} h={3.5} />
        </Box>
      </MobileNav>
    </>
  )
}

export default NavBar
