import { useRouter } from 'next/router'
import { useEffect, useState } from 'react'
import { cancelAll } from '../../services/methods'
import { DISCOVER_SIGNALS } from '../../utils/requests'
import { useSignalContext } from '../signal/context'
import { RouterQuery } from '../../globals'
import RequestHandler from '../../utils/handlers/requestHandler'

interface Exports {
  handleSearch: Function
  search: string
  filtered: any[]
  currentItems: any[]
  handlePageClick: Function
  pageCount: number
}

/**
 * Functional Handler Component, for handling all page logic
 *
 * @param {Function} setIsLoading loading setter from parent component
 * @returns {Exports} exported functions
 */
const PageLogic = (setIsLoading: Function): Exports => {
  const ITEMS_PER_PAGE = 10
  const router = useRouter()
  const { vesselid, page }: Partial<RouterQuery> = useRouter().query
  const { signalnames, set } = useSignalContext()
  const { request } = RequestHandler()
  const [search, setSearch] = useState<string>('')
  const [filtered, setFiltered] = useState<any[]>([])
  const [currentItems, setCurrentItems] = useState<any[] | null>(null)
  const [pageCount, setPageCount] = useState<number>(0)
  const [itemOffset, setItemOffset] = useState<number>(0)

  useEffect(() => {
    /**
     * Asynchronous function for discovering signals on render / vessel id change
     */
    const fetch = async () => {
      await request(DISCOVER_SIGNALS, {
        payload: {
          vesselid: vesselid as string
        },
        // eslint-disable-next-line jsdoc/require-jsdoc
        next: (data: any) => set._signalnames(data.signals),
        setIsLoading
      })
    }

    if (vesselid) fetch()

    return () => cancelAll()
  }, [vesselid])

  useEffect(() => {
    if (signalnames) setFiltered([...signalnames])
  }, [signalnames])

  useEffect((): any => {
    // Memory leak guard
    let isMounted = true

    const endOffset = itemOffset + ITEMS_PER_PAGE
    if (isMounted) setCurrentItems(filtered.slice(itemOffset, endOffset))
    if (isMounted) setPageCount(Math.ceil(filtered.length / ITEMS_PER_PAGE))

    // Clean-up
    return () => (isMounted = false)
  }, [itemOffset, filtered, page])

  /**
   * Helper Method for handling page change
   *
   * @param {any} e event
   */
  const handlePageClick = (e: any) => {
    router.push({ pathname: `/vessels/${vesselid}/page/${e.selected + 1}` })
    const offset = (e.selected * ITEMS_PER_PAGE) % signalnames.length
    setItemOffset(offset)
  }

  /**
   * Method for handling filtering for search key
   *
   * @param {string} key search key
   */
  const handleSearch = (key: string) => {
    setSearch(key)
    setFiltered(signalnames.filter((signal: string) => signal.toLowerCase().includes(key.toLowerCase())))
    handlePageClick({ selected: 0 })
  }

  return { handleSearch, search, filtered, currentItems, handlePageClick, pageCount }
}

export default PageLogic
