import { ChevronRightIcon } from '@chakra-ui/icons'
import { Badge, Box, Divider, Flex, Text, useColorModeValue } from '@chakra-ui/react'
import { useRouter } from 'next/router'
import { MouseEventHandler, ReactElement, useCallback } from 'react'
import QueryHandler from '../../utils/handlers/queryHandler'

interface Props {
  vesselid?: string
  currentItems: any[]
  small?: boolean
  onClick?: Function
}

/**
 * Functional Component for rendering List Item
 *
 * @param {Props} args item props
 * @returns {ReactElement<Props>} functional component
 */
const Items = ({ vesselid = '', currentItems, small = false, onClick }: Props): ReactElement<Props> => {
  const router = useRouter()
  const { signalid, compare } = router.query
  const { getCompare } = QueryHandler()

  /**
   * Helper Method for handling Item Click
   *
   * @param {any} item current item
   * @param {boolean} used check if item is in use, prevent action
   * @returns {MouseEventHandler<HTMLDivElement>} onClick
   */
  const handleClick = (item: any, used?: boolean): MouseEventHandler<HTMLDivElement> => {
    return () => {
      if (!small) router.push({ pathname: `/vessels/${vesselid}/signal/${item}`, query: { compare: 'null', from: '0', to: '0' } })
      if (onClick && !used) return onClick(item)
    }
  }

  /**
   * Helper method for checking if signal list item is already in use
   *
   * @returns {boolean} isUsed
   */
  const isUsed = useCallback((signal: string): boolean => {
    const comparison = getCompare()
    const list = [signalid, ...comparison]
    let result = false

    list.forEach((name: string) => {
      if (name === signal) result = true
    })

    return result
  }, [signalid, compare])

  return (
    <>
      {currentItems &&
        currentItems.map((item: any, idx: number) => (
          <span key={item}>
            {small && (
              <Box
                key={item}
                my={2}
                rounded="md"
                cursor="pointer"
                disabled={true}
                onClick={handleClick(item, isUsed(item))}
                _hover={{ bgColor: useColorModeValue('gray.100', 'gray.500') }}
                bgColor={useColorModeValue('gray.200', 'gray.600')}
              >
                <Flex justifyContent="space-between" alignItems="center">
                  <Text my={3} ml={2}>
                    {item}
                  </Text>
                  {isUsed(item) && <Badge colorScheme='green'>In Use</Badge>}
                  <ChevronRightIcon color="gray.500" mr={2} />
                </Flex>
              </Box>
            )}
            {!small && (
              <Box key={item} mr={6} cursor="pointer" onClick={handleClick(item)}>
                <Flex justifyContent="space-between" alignItems="center">
                  <Text my={3} ml={2}>
                    {item}
                  </Text>
                  <ChevronRightIcon color="gray.500" mr={2} />
                </Flex>
                {idx === currentItems.length - 1 ? null : <Divider />}
              </Box>
            )}
          </span>
        ))}
    </>
  )
}

export default Items
