import { Box, Flex, IconButton, useColorModeValue } from '@chakra-ui/react'
import { ReactElement, ReactNode, useEffect, useState } from 'react'
import { HiArrowSmLeft, HiArrowSmRight } from 'react-icons/hi'
import ReactPaginate from 'react-paginate'

interface Props {
  perPage?: number
  children: ReactNode
  items: any[]
  setItems: Function
  page: number
  totalLength: number
  onPageClick?: Function
}

/**
 * Functional Component for rendering Pagenation for List Components
 *
 * @param {Props} args pagenation props
 * @returns {ReactElement<Props>} functional component
 */
const Pagenation = ({ children, perPage = 8, items, setItems, page, onPageClick, totalLength }: Props): ReactElement<Props> => {
  const [pageCount, setPageCount] = useState<number>(0)
  const [itemOffset, setItemOffset] = useState<number>(0)

  useEffect(() => {
    const endOffset = itemOffset + perPage
    setItems(items.slice(itemOffset, endOffset))
    setPageCount(Math.ceil(items.length / perPage))
  }, [itemOffset, items])

  /**
   * Helper Method for handling Page click
   *
   * @param {any} e event
   */
  const handlePageClick = (e: any) => {
    if (onPageClick) onPageClick()
    const newOffset = (e.selected * perPage) % totalLength
    setItemOffset(newOffset)
  }

  return (
    <Box>
      {children}
      <Flex>
        <ReactPaginate
          nextLabel={<IconButton className="next" aria-label="next" icon={<HiArrowSmRight />} />}
          breakLabel={'...'}
          forcePage={page}
          className="pagenation"
          pageClassName="pagenation-list"
          onPageChange={handlePageClick}
          activeClassName={useColorModeValue('active-dark', 'active-light')}
          marginPagesDisplayed={2}
          pageRangeDisplayed={3}
          pageCount={pageCount}
          previousLabel={<IconButton className="previous" aria-label="previous" icon={<HiArrowSmLeft />} />}
        />
      </Flex>
    </Box>
  )
}

export default Pagenation
