import React, { ReactElement, ReactNode } from 'react'
import { Box, Text, Stack, Flex, useColorModeValue } from '@chakra-ui/react'
import NavLink from './navlink'
import Link from 'next/link'
import { AddIcon, CloseIcon } from '@chakra-ui/icons'
import VesselModal from './vesselmodal'
import { useSiteState } from '../../utils/context/state'
import { useRouter } from 'next/router'
import NavigationHandler from '../../utils/handlers/navigationHandler'

interface Props {
  children: ReactNode
}

/**
 * Functional Component for rendering page sidebar
 *
 * @param {Props} args sidebar props
 * @returns {ReactElement<Props>} functional component
 */
const SideBar = ({ children }: Props): ReactElement<Props> => {
  const { asPath } = useRouter()
  const { vessels, set } = useSiteState()
  const { isActive, getPathPrefix, handleRemoveVessel, getTitle } = NavigationHandler()

  /**
   * Helper Method for generating all side items
   *
   * @returns {ReactElement} component
   */
  const buildSideItems = (): ReactElement => {
    if (!vessels) return null
    if (vessels.length === 0) return null
    return vessels.map((vessel: string, index: number) => (
      <Box key={index} display="flex" alignItems="center">
        <NavLink key={vessel + index} active={isActive(vessel)} activeColor={useColorModeValue('gray.200', 'gray.700')} href={`${getPathPrefix(asPath)}/${vessel}/page/1`}>
          {vessel || 'Loading...'}
        </NavLink>
        <CloseIcon _hover={{ transform: 'scale(1.5)' }} ml={3} mr={2.5} cursor="pointer" transition="ease .2s" onClick={() => handleRemoveVessel(vessel)} w={3} h={3} />
      </Box>
    ))
  }

  return (
    <Flex justifyContent="flex-start" flex="1" h="full" transition="all ease .2s" py={{ base: 5, lg: 10 }}>
      {asPath.includes('vessels') && (
        <Box px={2} display={{ base: 'none', md: 'block' }} w={60}>
          <Box zIndex={5} backdropFilter="blur(5px)" mb={5}>
            <Link href={`/${getTitle(asPath).toLowerCase()}`}>
              <Text cursor="pointer" ml={2} fontSize="md" fontWeight="bold" opacity={0.8}>
                {getTitle(asPath)}
              </Text>
            </Link>
          </Box>

          <Stack as={'nav'} spacing={2} pr={2}>
            {buildSideItems()}
          </Stack>
          <Box
            mt={5}
            px={2}
            py={1}
            cursor="pointer"
            outline="0 0 0 3px var(--chakra-colors-gray-500)"
            rounded={'md'}
            opacity={0.8}
            fontWeight={600}
            bg={useColorModeValue('gray.200', 'gray.700')}
            _hover={{
              bg: useColorModeValue('gray.300', 'gray.600')
            }}
            display="flex"
            alignItems="center"
            justifyContent="space-between"
            onClick={set.onOpen}
          >
            Add
            <AddIcon color="green.300" w={3.5} h={3.5} />
          </Box>
        </Box>
      )}
      <Flex flex="1">{children}</Flex>
      <VesselModal />
    </Flex>
  )
}

export default SideBar
