import { ChevronDownIcon } from '@chakra-ui/icons'
import { Wrap, Box, useColorModeValue, Menu, Text, MenuButton, MenuList, Button, Divider, Flex, CloseButton, Stack, IconButton, InputGroup, InputRightAddon, Input } from '@chakra-ui/react'
import { useRouter } from 'next/router'
import { ReactElement, useState } from 'react'
import { DayPicker } from 'react-day-picker'
import { HiOutlineArrowsExpand, HiPlus } from 'react-icons/hi'
import { BiCollapse, BiTime } from 'react-icons/bi'
import { colors } from '../../utils/constants'
import { getFormattedDate, getHourMinute, normalizeNumber, parse } from '../../utils/tools'
import { useSignalContext } from './context'
import SignalSettings from './settings'
import QueryHandler from '../../utils/handlers/queryHandler'
import { FETCH_RANGE } from '../../utils/requests'
import RequestHandler from '../../utils/handlers/requestHandler'
import { HandledResponse, RouterQuery, SuccessResponse } from '../../globals'
import RangeResponseHandler from '../../utils/handlers/rangeResponseHandler'

/**
 * Functional Component for rendering all signal controls
 *
 * @returns {ReactElement} functional component
 */
const Controls = (): ReactElement => {
  const { vesselid, signalid }: Partial<RouterQuery> = useRouter().query
  const { isLoading, expanded, set, chartRef } = useSignalContext()
  const now = Date.now()
  const h = normalizeNumber(new Date(now).getHours())
  const m = normalizeNumber(new Date(now).getMinutes())
  const [date, setDate] = useState<number>(now)
  const { request } = RequestHandler()
  const [time, setTime] = useState<string>(`${h}:${m}`)
  const [offset, setOffset] = useState<number | string>(5)
  const { getCompare, setCompare } = QueryHandler()
  const { handleRangeSuccess, handleRangeFailure } = RangeResponseHandler()

  /**
   * Helper Method for Picking Date
   *
   * @param {any} day picked day
   */
  const handleDatePick = (day: any): void => {
    setDate(new Date(day).getTime())
  }

  /**
   * Helper Method for getting number of offset
   *
   * @param {string} _offset input
   * @returns {number} output
   */
  const getOffset = (_offset: number | string): number => {
    if (typeof _offset === 'string') return parseInt(_offset)
    return _offset
  }

  /**
   * Method for Sending Range Request on first render of signal page
   */
  const handleSendRange = async (): Promise<void> => {
    const { h, m } = getHourMinute(time.replace(':', ''))
    const from = new Date(date).setHours(h, m, 0, 0)
    const to = from + getOffset(offset) * 60 * 1e3
    const comparison = getCompare()
    const start = performance.now()

    // Manually zoom to area of interest before manual fetch
    if (chartRef.current) chartRef.current.zoomScale('x', { min: from, max: to }, 'none')

    await request(FETCH_RANGE, {
      payload: {
        vesselid: vesselid as string,
        from,
        to,
        signals: [signalid as string, ...comparison]
      },
      // eslint-disable-next-line jsdoc/require-jsdoc
      next: (data: SuccessResponse) => handleRangeSuccess(data, { min: from, max: to, start }),
      // eslint-disable-next-line jsdoc/require-jsdoc
      error: (data: HandledResponse, status: number) => handleRangeFailure(data, status),
      toast: false, // Since we manually want to tell the user about the status of the fetch, we disable automatic toasts
      setIsLoading: set._isLoading
    })
  }

  /**
   * Handle remove signal from comparison
   *
   * @param {string} signal signal name
   */
  const handleRemove = (signal: string): void => {
    const comparison = getCompare()
    const filtered = comparison.filter((_signal: string) => _signal !== signal)
    setCompare(filtered)
  }

  /**
   * Helper Method for handling change from time-input
   *
   * @param {any} e event
   */
  const handleChange = (e: any): void => {
    setTime(e.target.value)
  }

  /**
   * Helper method for toggling in and out of expanded view mode
   */
  const handleExpand = (): void => {
    set._expanded(!expanded)
  }

  return (
    <Box w="100%" px={5} py={3} mb={4} rounded="md" shadow={useColorModeValue('sm', 'md')} bgColor={useColorModeValue('white', 'gray.700')}>
      <Flex>
        <Flex gridGap={3} flexDir={{ base: 'column', lg: 'row' }}>
          <Flex gridGap={3}>
            <IconButton aria-label="refresh" onClick={handleExpand} icon={!expanded ? <HiOutlineArrowsExpand /> : <BiCollapse />} />
            <Menu>
              <MenuButton as={Button} w="18rem" rightIcon={<ChevronDownIcon />}>
                {getFormattedDate(date)}
              </MenuButton>
              <MenuList zIndex={1000}>
                <DayPicker
                  mode="single"
                  selected={new Date(date)}
                  onSelect={handleDatePick}
                  modifiersClassNames={{
                    selected: 'selected',
                    today: 'today'
                  }}
                />
              </MenuList>
            </Menu>
          </Flex>
          <Flex flexDir={{ base: 'column', lg: 'row' }} gridGap={3}>
            <Flex gridGap={3}>
            <InputGroup variant="filled">
              <Input textAlign="center" placeholder="10:00" onChange={handleChange} value={time} />
              <InputRightAddon children={<BiTime />} />
            </InputGroup>
            <InputGroup variant="filled">
              <Input textAlign="center" placeholder="5" onChange={(e: any) => setOffset(parse(e.target.value))} value={offset} max={120} />
              <InputRightAddon children="Offset" />
            </InputGroup>
            </Flex>
            <Button onClick={handleSendRange} isLoading={isLoading}>
              Send
            </Button>
          </Flex>
        </Flex>
      </Flex>
      <Divider my={4} />
      <Flex>
        <Stack direction="row" gridGap={3}>
          <Flex h={10} bgColor={useColorModeValue('gray.100', 'rgba(211, 211, 211, 0.1)')} rounded="md">
            <Box ml={2} my="auto" bgColor={colors[0]} h="3" w="3" borderRadius="lg"></Box>
            <Text my="auto" px={2}>
              {signalid}
            </Text>
            <SignalSettings signalid={signalid} />
          </Flex>
          <Divider orientation="vertical" />
          <Wrap spacing={4}>
            {getCompare().map((_signalid: string, idx: number) => (
              <Flex h={10} key={idx} bgColor={useColorModeValue('gray.100', 'rgba(211, 211, 211, 0.1)')} rounded="md">
                <Box ml={2} my="auto" bgColor={colors[idx + 1]} h="3" w="3" borderRadius="lg"></Box>
                <Text my="auto" px={2}>
                  {_signalid}
                </Text>
                <SignalSettings signalid={_signalid} />
                <CloseButton disabled={isLoading} onClick={() => handleRemove(_signalid)} my="auto" size="sm" mr={2} />
              </Flex>
            ))}
            <IconButton name="modal-button" disabled={getCompare().length > 4 || isLoading} aria-label="Search database" onClick={set.onOpen} icon={<HiPlus />} />
          </Wrap>
        </Stack>
      </Flex>
    </Box>
  )
}

export default Controls
