import { ChevronDownIcon } from '@chakra-ui/icons'
import { Text, IconButton, Popover, PopoverContent, PopoverTrigger, useDisclosure, useOutsideClick, Box, PopoverHeader, Flex, Stack, Switch, Slider, SliderTrack, SliderFilledTrack, SliderThumb } from '@chakra-ui/react'
import { ReactElement, useEffect, useRef, useState } from 'react'
import { useSignalContext } from './context'

interface Props {
  signalid: string | string[] | undefined
}

/**
 * Functional Component for rendering signal settings
 *
 * @param {Props} args settings props
 * @returns {ReactElement<Props>} functional component
 */
const SignalSettings = ({ signalid }: Props): ReactElement<Props> => {
  const [tension, setTension] = useState<number>(0.4)
  const [radius, setRadius] = useState<number>(0)
  const [line, setLine] = useState<number>(2)
  const [stepped, setStepped] = useState<string | boolean>('middle')
  const ref = useRef()
  const { chartRef } = useSignalContext()
  const { onOpen, onClose, isOpen } = useDisclosure()
  useOutsideClick({
    ref,
    // eslint-disable-next-line jsdoc/require-jsdoc
    handler: () => onClose()
  })

  /**
   * Helper for toggling Stepped Option
   *
   * @param {any} e event
   */
  const handleSteppedToggle = (e: any): void => {
    if (e.target.checked) {
      setStepped('middle')
      return
    }
    setStepped(false)
  }

  useEffect(() => {
    const chart = chartRef.current
    if (chart) {
      const datasets = chart.data.datasets

      const match = datasets.find((dataset: any) => dataset.label === signalid)
      if (match) {
        match.tension = tension
        match.radius = radius
        match.borderWidth = line
        match.stepped = stepped

        chart.data.datasets = [...datasets]

        chart.stop()
        chart.update()
      }
    }
  }, [tension, radius, line, stepped])

  return (
    <Popover isOpen={isOpen} onOpen={onOpen} onClose={onClose} placement="bottom" closeOnBlur={false}>
      <PopoverTrigger>
        <IconButton aria-label='button' icon={<ChevronDownIcon />} size="xs" my="auto" variant="ghost" mr={2} />
      </PopoverTrigger>
      <PopoverContent w="23rem">
        <Box ref={ref}>
          <PopoverHeader>Line Settings - {signalid || 'MISSING_SIGNALID'}</PopoverHeader>
          <Flex gridGap={4} p={5}>
            <Stack direction="row" w="5rem">
              <Text fontWeight={600} my="auto" fontSize={13}>
                Stepped
              </Text>
              <Box display="flex" mr={4}>
                <Switch h={5} my="auto" onChange={handleSteppedToggle} sx={{ transform: 'rotate(90deg)' }} defaultChecked checked={stepped === 'middle'} colorScheme="green" size="md" />
              </Box>
            </Stack>
            <Stack direction="column" w="6rem">
              <Text fontWeight={600} my="auto" fontSize={13}>
                {`Radius ${radius}`}
              </Text>
              <Slider size="sm" value={radius} min={0} max={5} step={1} onChange={(val: any) => setRadius(val)}>
                <SliderTrack>
                  <SliderFilledTrack bg="green.300" />
                </SliderTrack>
                <SliderThumb />
              </Slider>
            </Stack>
            <Stack direction="column" w="5rem">
              <Text fontWeight={600} my="auto" fontSize={13}>
                {`Line ${line}`}
              </Text>
              <Slider size="sm" value={line} min={1} max={6} step={1} onChange={(val: any) => setLine(val)}>
                <SliderTrack>
                  <SliderFilledTrack bg="green.300" />
                </SliderTrack>
                <SliderThumb />
              </Slider>
            </Stack>
            <Stack direction="column" justifyContent="space-between" w="6rem" opacity={stepped === 'middle' ? 0.3 : 1}>
              <Text fontWeight={600} my="auto" fontSize={13}>
                {`Tension ${tension}`}
              </Text>
              <Slider size="sm" disabled={stepped === 'middle'} value={tension} min={0} max={1} step={0.1} onChange={(val: any) => setTension(val)}>
                <SliderTrack>
                  <SliderFilledTrack bg="green.300" />
                </SliderTrack>
                <SliderThumb />
              </Slider>
            </Stack>
          </Flex>
        </Box>
      </PopoverContent>
    </Popover>
  )
}

export default SignalSettings
