import { Box, Text, Button } from '@chakra-ui/react'
import { useRouter } from 'next/router'
import { useEffect } from 'react'
import { ToastStatus } from '../../utils/constants'
import { DISCOVER_SIGNALS, FETCH_RANGE } from '../../utils/requests'
import ToastHandler from '../../utils/handlers/toastHandler'
import { debounce, getSignalNames } from '../../utils/tools'
import { useSignalContext } from './context'
import { cancelAll } from '../../services/methods'
import QueryHandler from '../../utils/handlers/queryHandler'
import RequestHandler from '../../utils/handlers/requestHandler'
import { FetchResponse, HandledResponse, RouterQuery, Signal } from '../../globals'
import RangeResponseHandler from '../../utils/handlers/rangeResponseHandler'
import ChartDataHandler from '../../utils/handlers/chartdataHandler'

interface Exports {
  handleAdd: Function
  handleFetch: Function
}

/**
 * Functional Component that handles all logic of Signal Page
 *
 * @returns {Exports} functional component
 */
const SignalLogic = (): Exports => {
  const { signalnames, set, chartRef, isLoading, signals } = useSignalContext()
  const { vesselid, signalid, compare }: Partial<RouterQuery> = useRouter().query
  const query: Partial<RouterQuery> = useRouter().query
  const { request } = RequestHandler()
  const { pendingToast, updateToast } = ToastHandler()
  const { getCompare, setCompare, setRange, getRange } = QueryHandler()
  const { handleRangeSuccess, handleRangeFailure } = RangeResponseHandler()
  const { findMinMax, hydrateDatasets, removeDataset, handleDatasetColors } = ChartDataHandler()

  useEffect(() => {
    /**
     * Asynchronous fetch method for getting data of signals on-render
     */
    const fetch = async (): Promise<void> => {
      const comparison = getCompare()
      // Have to go back in time, as cache API is running behing in the cloud
      let _to = Date.now() - 1.5e4
      let _from = _to - 1e5

      // If range is found in URL, make this the starting point
      if (query.from && query.to) {
        if (query.from !== '0' || query.to !== '0') {
          const { from, to } = getRange()
          _from = from
          _to = to
        } else {
          setRange({ from: _from, to: _to })
        }
      } else {
        setRange({ from: _from, to: _to })
      }

      await request(FETCH_RANGE, {
        payload: {
          vesselid: vesselid as string,
          from: _from,
          to: _to,
          signals: [signalid, ...comparison] as string[]
        },
        // eslint-disable-next-line jsdoc/require-jsdoc
        next: (data: FetchResponse) => {
          set._signals(data.signal)
          hydrateDatasets(data.signal)
        },
        setIsLoading: set._isLoading
      })
    }

    if (vesselid) fetch()

    return () => cancelAll()
  }, [vesselid])

  useEffect(() => {
    /**
     * Asynchronous fetch method for getting signalnames if missing
     */
    const fetch = async () => {
      await request(DISCOVER_SIGNALS, {
        payload: { vesselid: vesselid as string },
        // eslint-disable-next-line jsdoc/require-jsdoc
        next: (data: any) => {
          set._signalnames(data.signals)
        }
      })
    }

    // Only request these signals if "signalnames" is empty, thus preventing double fetching
    if ((!signalnames || signalnames.length === 0) && vesselid) fetch()

    return () => cancelAll()
  }, [vesselid])

  useEffect(() => {
    const comparison = getCompare()
    if (!signals) return
    if (signals.length === 0 || comparison.length === signals.length - 1) return
    // TODO: remove data set that is only found in signals
    if (comparison.length < signals.length - 1) {
      let result: Signal[] = []

      let name: null | string = null

      signals.forEach((s: Signal) => {
        if (!comparison.toString().includes(s.name)) name = s.name
      })

      // Remove dataset from chart
      removeDataset(name)
      // Re-arrange colors if needed
      handleDatasetColors()

      result = signals.filter((s: Signal) => {
        if (s.name === signalid) return true
        return s.name !== name
      })

      set._signals(result)
    }
  }, [compare])

  useEffect(() => {
    let timeout: any
    let abort: any

    if (isLoading) {
      timeout = setTimeout(() => {
        pendingToast('Fetching still processing...', 'The request is taking longer than usual')
      }, 7000)
      abort = setTimeout(() => {
        updateToast(
          ToastStatus.Info,
          'Fetching still processing...',
          <Box display="flex" flexDir="column" justifyContent="center">
            <Text>Fetching has surpassed 15 seconds</Text>
            <Button onClick={() => cancelAll()} colorScheme="black" variant="outline" mt={4}>
              Abort Fetch
            </Button>
          </Box>
        )
      }, 15000)
    }

    if (!isLoading) {
      clearTimeout(timeout)
      clearTimeout(abort)
    }

    return () => {
      clearTimeout(timeout)
      clearTimeout(abort)
    }
  }, [isLoading])

  /**
   * Method for handling the addition of new signal to comparison
   *
   * @param {string} signal signal name
   */
  const handleAdd = async (signal: string): Promise<void> => {
    set.onClose()
    try {
      const comparison = getCompare()
      if (!comparison.includes(signal) && signal !== signalid) {
        setCompare([...comparison, signal])
        if (!chartRef.current) return
        const { min, max } = findMinMax(chartRef.current.data.datasets)
        await request(FETCH_RANGE, {
          payload: {
            vesselid: vesselid as string,
            from: Math.round(min),
            to: Math.round(max),
            signals: [...getSignalNames(signals), signal] as string[]
          },
          // eslint-disable-next-line jsdoc/require-jsdoc
          next: (data: FetchResponse) => {
            set._signals(data.signal)
            hydrateDatasets(data.signal)
          },
          setIsLoading: set._isLoading
        })
      }
    } catch (err: any) {
      console.warn(err)
    }
  }

  // Fetch Method that is executes after 1000 ms after last attempt at calling
  const fetch = debounce(async (setIsLoading: Function) => {
    if (!chartRef.current) return
    const { min, max } = chartRef.current.scales.x

    const comparison = getCompare()
    const start = performance.now()

    await request(FETCH_RANGE, {
      payload: {
        vesselid: vesselid as string,
        from: Math.round(min),
        to: Math.round(max),
        signals: [signalid as string, ...comparison]
      },
      // eslint-disable-next-line jsdoc/require-jsdoc
      next: (data: FetchResponse) => handleRangeSuccess(data, { min, max, start }),
      // eslint-disable-next-line jsdoc/require-jsdoc
      error: (data: HandledResponse, status: number) => handleRangeFailure(data, status),
      toast: false, // Since we manually want to tell the user about the status of the fetch, we disable automatic toasts
      setIsLoading
    })
  }, 1e3)

  /**
   * Handle Fetch method checks inputs and fetches new range when appropriate
   *
   * @param {any} limit min, max
   * @param {number} limit.min min
   * @param {number} limit.max min
   * @param {any} chart chart instance
   */
  const handleFetch = (limit: { min: number; max: number }, chart: any): void => {
    try {
      const { min, max } = chartRef.current.scales.x
      // Set path query range for persisting chart state through URL
      setRange({ from: Math.round(min), to: Math.round(max) })
      // Prevent fetching same data tha we already have while interaction with chart
      const difference = Math.abs(min - limit.min + max - limit.max)
      // TODO: If (difference < MIN_MOVEMENT_LIMIT) return;
      if (difference === 0) return

      // Normal behavior
      set._isLoading(true)
      // Debounces call
      fetch(set._isLoading)
    } catch (err: any) {
      console.warn('PRE_FETCH_ERROR', err)
    }
  }
  return { handleAdd, handleFetch }
}

export default SignalLogic
