import { LTTB, DataPoint, Indexable } from 'downsample'
import { useSignalContext } from '../../components/signal/context'
import { Dataset, Signal } from '../../globals'
import { colors } from '../constants'
import { sortDoubles } from '../tools'

const STEPPED = 'middle'
const TENSION = 0.4
const RADIUS = 0
const LINE = 2

const datasetbase: Partial<Dataset> = {
  type: 'line' as const,
  borderWidth: LINE,
  indexAxis: 'x',
  radius: RADIUS,
  fill: true,
  stepped: STEPPED,
  tension: TENSION,
  parsing: false,
  spanGaps: true
}

interface Exports {
  hydrateDatasets: Function
  findMinMax: Function
  createDataset: Function
  removeDataset: Function
  handleDatasetColors: Function
}

/**
 * Functional Handler Component, for handling all chart data related actions
 *
 * @returns {Exports} exported functions
 */
const ChartDataHandler = (): Exports => {
  const { chartRef } = useSignalContext()

  /**
   * Handler for edge-case: only 1 point of data, make 2 extra points for min and max of chart x-scale
   *
   * @param {Indexable<DataPoint>} data downsampled data
   * @returns {Indexable<DataPoint>} handled data
   */
  const handleSingleDouble = (data: Indexable<DataPoint>): Indexable<DataPoint> => {
    if (data.length > 1) return data

    if (chartRef.current) {
      const { min, max } = chartRef.current.scales.x
      const value: { x: number, y: number } = data[0] as any

      if (value) {
        data[0] = ({ y: value.y, x: min })
        data[1] = ({ y: value.y, x: value.x })
        data[2] = ({ y: value.y, x: max })
      }
    }

    return data
  }

  /**
   * Method for sorting, formatting, and downsampling signal data
   *
   * @param {Signal} signal signals from network request
   * @param {number} length signals length
   * @returns {Indexable<DataPoint>} indexable list of x and y points
   */
  const downsample = (signal: Signal, length?: number): Indexable<DataPoint> => {
    // Sort unix timestamps to achieve ordinal data for our time-series chart
    const sorted = sortDoubles(signal.double)
    // Format the data to be a list of x and y points
    const formatted = sorted.map((key: string) => ({ y: signal.double[key], x: parseInt(key) }))
    // Grab width of chart area to decide how much we should downsample the data
    const width = chartRef.current ? chartRef.current.chartArea.width : 500
    // Grab count of datasets in our chart to make downsample factor greater for more datasets
    let datasets: number = chartRef.current ? chartRef.current.data.datasets.length : 1
    if (length) datasets = length
    // We want to downsample based on the length of our chart and the count of our datasets
    const downsampled = LTTB(formatted, width / (datasets + 1))

    return handleSingleDouble(downsampled)
  }

  /**
   * Method for creating a custom ChartJS dataset
   *
   * @param {Signal} signal signal data
   * @param {number} index decides color and connects to y-scale of same index
   * @param {number} length length of dataset
   * @returns {Partial<Dataset>} generated dataset
   */
  const createDataset = (signal: Signal, index: number, length?: number): Partial<Dataset> => {
    return {
      ...datasetbase,
      label: signal.name,
      yAxisID: signal.name,
      data: downsample(signal, length),
      backgroundColor: colors[index],
      borderColor: colors[index]
    }
  }

  /**
   * Helper Method for re-hydrating the chart with new data without default ChartJS updated "update("none")"
   *
   * @param {Signal[]} signals list of signals from network call
   */
  const hydrateDatasets = (signals: Signal[]): void => {
    if (!chartRef.current) return
    const chart = chartRef.current

    const datasets: any = []

    if (signals.length > 0) {
      signals.forEach((signal: Signal) => {
        // TODO: fix edge case when dataset is not found, set correct color based on new index, if dataset is missing it will get same color as another dataset
        if (Object.keys(signal.double).length > 0) {
          const index = chart.data.datasets.findIndex((entry: any) => entry.label === signal.name)
          let dataset: Partial<Dataset> | null = chart.data.datasets[index]

          // We want the performance of the LTTB algorithm but calculating it with the width / n, given n signals
          const downsampled = downsample(signal, signals.length)

          if (!dataset) {
            const length = datasets.length
            dataset = createDataset(signal, length, signals.length)
          }

          datasets.push({
            ...dataset,
            data: downsampled
          })
        }
      })

      chart.data.datasets = [...datasets]

      chart.stop()
      chart.update('none')
    }
  }

  /**
   * Helper Method for removing dataset by name from chart instance
   *
   * @param {string} name dataset to be deleted
   */
  const removeDataset = (name: string): void => {
    const chart = chartRef.current
    if (chart) {
      const datasets = chart.data.datasets.filter((entry: any) => entry.label !== name)

      chart.data.datasets = [...datasets]

      chart.stop()
      chart.update('none')
    }
  }

  /**
   * Helper method for re-arranging colors after a comparison change
   *
   * @returns {void} nothing
   */
  const handleDatasetColors = (): void => {
    if (!chartRef.current) return
    const chart = chartRef.current

    chart.data.datasets.forEach((dataset: Dataset, idx: number) => {
      dataset.backgroundColor = colors[idx]
      dataset.borderColor = colors[idx]
    })

    chart.stop()
    chart.update('none')
  }

  /**
   * Helper Method for finding min and max x value of all datasets
   *
   * @param {Partial<Dataset>[]} datasets available datasets
   * @returns {object} { min: number, max: number }
   */
  const findMinMax = (datasets: any): { min: number; max: number } => {
    const values = { min: -1, max: -1 }
    datasets.forEach((dataset: any) => {
      try {
        if (values.min < 0) {
          values.min = dataset.data.at(0).x
        } else {
          if (values.min > dataset.data.at(0).x) values.min = dataset.data.at(0).x
        }
        if (values.max < 0) {
          values.max = dataset.data.at(-1).x
        } else {
          if (values.max < dataset.data.at(-1).x) values.max = dataset.data.at(-1).x
        }
      } catch (err) {
        console.warn('Could not find data min max')
        return values
      }
    })

    return values
  }

  return { hydrateDatasets, findMinMax, createDataset, removeDataset, handleDatasetColors }
}

export default ChartDataHandler
