import { useRouter } from 'next/router'
import { navigation } from '../constants'
import { useSiteState } from '../context/state'
import { capitalize } from '../tools'
import ToastHandler from './toastHandler'

interface Exports {
    isActive: Function
    getNavItem: Function
    getTitle: Function
    getPathPrefix: Function
    handleRemoveVessel: Function
}

/**
 * Functional Component for handling all navigation actions from sidebar or mobile nav
 *
 * @returns {Exports} functional component
 */
const NavigationHandler = (): Exports => {
  const router = useRouter()
  const { set } = useSiteState()
  const { successToast } = ToastHandler()
  const { vesselid } = router.query

  /**
   * Helper method for checking if a nav link is active
   *
   * @param {string} id vessel id
   * @returns {boolean} isActive
   */
  const isActive = (id: string): boolean => {
    if (!vesselid) return false
    return vesselid === id
  }

  /**
   * Helper method for getting nav item based on path
   *
   * @param {string} path given path
   * @returns {any} navitem
   */
  const getNavItem = (path: string | undefined): any => {
    if (!path) return
    const navitem = navigation.find((nav: any) => path.includes(nav.title.toLowerCase()))
    if (!navitem) return
    return navitem
  }

  /**
   * Helper Method for getting title of nav item
   *
   * @param {string} path given path
   * @returns {string} title
   */
  const getTitle = (path: string | undefined): string => {
    const navitem = getNavItem(path)
    if (!navitem) return 'PATH_NOT_FOUND'
    return capitalize(navitem.title)
  }

  /**
   * Helper Method for getting prefix of nav item path
   *
   * @param {string} path given path
   * @returns {string} nav item path
   */
  const getPathPrefix = (path: string | undefined): string => {
    const navitem = getNavItem(path)
    if (!navitem) return ''
    return navitem.path
  }

  /**
   * Method for handling removal of a vessel
   *
   * @param {string} vesselid vessel to be removed
   */
  const handleRemoveVessel = (vesselid: string) => {
    set._vessels((v: string[]) => [...v.filter((_v: string) => _v !== vesselid)])
    successToast('Removed Successfully', `Vessel '${vesselid}' was removed!`)
    router.push({ pathname: '/vessels' })
  }

  return { isActive, getNavItem, getTitle, getPathPrefix, handleRemoveVessel }
}

export default NavigationHandler
