import { useRouter } from 'next/router'
import { RouterQuery } from '../../globals'

interface Exports {
  setCompare: Function
  getCompare: Function
  setRange: Function
  getRange: Function
}

/**
 * Functional Handler Component, for handling all query related actions
 *
 * @returns {Exports} exported functions
 */
const QueryHandler = (): Exports => {
  const router = useRouter()
  const { compare, from, to }: Partial<RouterQuery> = useRouter().query

  /**
   * General Helper Method for setting any query params
   *
   * @param {any} args object of query keys
   */
  const setParams = (args: any): void => {
    const pathname = router.asPath.split('?')[0]
    router.push(
      {
        pathname,
        query: {
          compare,
          from: from || '0',
          to: to || '0',
          ...args
        }
      },
      undefined,
      { shallow: true }
    )
  }

  /**
   * Setter for Range
   *
   * @param {any} args unix timestamps
   */
  const setRange = ({ from, to }: { from: number; to: number }): void => {
    setParams({ from, to })
  }

  /**
   * Accessor for Range
   *
   * @returns {object} range
   */
  const getRange = (): { from: number; to: number } => {
    if (!from || !to) return { from: 0, to: 0 }
    try {
      const _from = parseInt(from as string)
      const _to = parseInt(to as string)
      if (!isNaN(_from) || !isNaN(_to)) return { from: _from, to: _to }
      return { from: 0, to: 0 }
    } catch (err: any) {
      console.warn('RANGE_PARAM_ERROR', err)
      return { from: 0, to: 0 }
    }
  }

  /**
   * Setter for Compare
   *
   * @param {string[]} signals signalnames
   */
  const setCompare = (signals: string[]): void => {
    const data = signals.toString().replaceAll(',', ' ')
    setParams({ compare: data || 'null' })
  }

  /**
   * Accessor for Compare
   *
   * @returns {object} compare
   */
  const getCompare = (): string[] => {
    if (!compare) return []
    try {
      if (compare === 'null') return []
      if (Array.isArray(compare)) return compare
      const _signals = compare.split(' ')
      return _signals
    } catch (err: any) {
      console.warn('COMPARE_PARAM_ERROR', err)
      return []
    }
  }

  return { setCompare, getCompare, setRange, getRange }
}

export default QueryHandler
