import { FetchResponse, HandledResponse } from '../../globals'
import { ToastStatus } from '../constants'
import { FETCH_RANGE } from '../requests'
import { titleOf } from '../tools'
import ChartDataHandler from './chartdataHandler'
import ToastHandler from './toastHandler'

interface Exports {
    handleRangeSuccess: Function
    handleRangeFailure: Function
}

/**
 * Functional Component for handling all extended range responses
 *
 * @returns {Exports} functional component
 */
const RangeResponseHandler = (): Exports => {
  const { resolveToast, warningToast, errorToast } = ToastHandler()
  const { hydrateDatasets } = ChartDataHandler()

  /**
   * Helper Method for handling success of Range Fetch
   *
   * @param {FetchResponse} data fetch response
   * @param {any} args min, max, start
   */
  const handleRangeSuccess = (data: FetchResponse, { min, max, start }: any): void => {
    const end = performance.now()
    // Tell user time of request
    resolveToast(ToastStatus.Success, 'Fetching Completed!', `Fetching took ${((end - start) / 1e3).toFixed(3)} seconds`)
    // Push signals into chart dataset
    hydrateDatasets(data.signal)
  }

  /**
   * Helper Method for handling failure of Range Fetch
   *
   * @param {HandledResponse} data handled error response
   * @param {number} status HTTPS status
   * @returns {void}
   */
  const handleRangeFailure = (data: HandledResponse, status: number): void => {
    const title = titleOf(ToastStatus.Error, FETCH_RANGE) + ` Status: ${status}`
    if (status > 499) {
      if (typeof data === 'string') return errorToast(title, `${data}`)
      if ('message' in data) return errorToast(title, `${data.message}`)
    } else {
      if (typeof data === 'string') return warningToast(title, `${data}`)
      if ('message' in data) return warningToast(title, `${data.message}`)
    }
    resolveToast(ToastStatus.Error, title, `${data}`)
  }
  return { handleRangeSuccess, handleRangeFailure }
}

export default RangeResponseHandler
