import { DiscoverPayload, HandledResponse, LastNPayload, RangePayload, SuccessResponse } from '../../globals'
import { ToastStatus } from '../constants'
import { useSiteState } from '../context/state'
import { titleOf } from '../tools'
import { requestReducer } from './requestReducer'
import ToastHandler from './toastHandler'

interface Exports {
  request: Function
}

/**
 * Functional Handler Component, for handling all network request related actions
 *
 * @returns {Exports} exported functions
 */
const RequestHandler = (): Exports => {
  const { set } = useSiteState()
  const { errorToast, infoToast, warningToast } = ToastHandler()

  interface RequestProps {
    payload: Partial<DiscoverPayload | LastNPayload | RangePayload>
    setIsLoading?: Function
    next?: Function
    error?: Function
    toast?: boolean
  }

  /**
   * Method to access all network calls, also handles side effects
   *
   * @param {string} key network call key
   * @param {any} args request arguments
   * @returns {Promise<HandledResponse | void>} response
   */
  const request = async (key: string, { payload, setIsLoading, next, error, toast = true }: RequestProps): Promise<HandledResponse | void> => {
    // If functions is added to arg, start loading
    if (setIsLoading) setIsLoading(true)

    const res: HandledResponse | void = await requestReducer(key, {
      ...payload,
      /**
       * Success callback
       *
       * @param {HandledResponse} data response
       * @param {number} status HTTP status
       * @returns {any} data
       */
      next: (data: HandledResponse, status: number) => {
        // edge-case "5 second timeout"
        if (typeof data === 'string') {
          handleFailure(key, data, toast, status)
          if (error) error(data, status)
          return
        }

        handleSuccess(key, data as any, toast)
        if (next) return next(data)
      },
      /**
       * Error callback
       *
       * @param {HandledResponse} data response
       * @param {number} status HTTP status
       * @returns {any} data
       */
      error: (data: HandledResponse, status: number) => {
        handleFailure(key, data, toast, status)
        if (error) return error(data, status)
      }
    })

    // If functions is added to arg, stop loading
    if (setIsLoading) setIsLoading(false)

    // Incase anyone wants the request to be handled in sync
    return res
  }

  /**
   * Helper method for handling success side-effect
   *
   * @param {string} key network call key
   * @param {SuccessResponse} data success response
   * @param {boolean} toast if true, toast will be displayed
   */
  const handleSuccess = (key: string, data: SuccessResponse, toast: boolean) => {
    if (data) {
      if ('bytecost' in data) set._bytecost((b: number) => b + data.bytecost)
      if ('bytetransfer' in data) set._bytetransfer((b: number) => b + data.bytetransfer)
    }
  }

  /**
   * Helper method for handling failure side-effects
   *
   * @param {string} key network call key
   * @param {HandledResponse} data success response
   * @param {boolean} toast if true, toast will be displayed
   * @param {number} status status from error response
   * @returns {void}
   */
  const handleFailure = (key: string, data: HandledResponse | string, toast: boolean, status: number): void => {
    if (toast) {
      // eslint-disable-next-line jsdoc/require-jsdoc
      const title = (_status: ToastStatus) => titleOf(_status, key) + ` Status: ${status}`
      if (typeof data === 'string') {
        if (data.includes('clean-up')) return infoToast(title(ToastStatus.Info), `${data}`)
        return warningToast(title(ToastStatus.Warning), `${data}`)
      }
      if (typeof data !== 'object') return errorToast(title(ToastStatus.Error), 'Please try again')
      if ('message' in data) return errorToast(title(ToastStatus.Error), `${data.message}`)
    }
  }

  return { request }
}

export default RequestHandler
