import { HandledResponse, NetworkResponse } from '../../globals'
import { wipe } from '../../services/commands'
import { getMetadata } from '../../services/metadata'
import { discover, getLastN, getRange } from '../../services/vessel'
import { DISCOVER_SIGNALS, FETCH_LASTN, FETCH_METADATA, FETCH_RANGE, WIPE_DATABASE } from '../requests'

/**
 * Map for connecting all network call keys with their methods
 */
const connections: { [key: string]: (payload: any) => Promise<HandledResponse | void> } = {
  // eslint-disable-next-line jsdoc/require-jsdoc
  [DISCOVER_SIGNALS]: (payload: any) => handleCall(discover, payload),
  // eslint-disable-next-line jsdoc/require-jsdoc
  [FETCH_LASTN]: (payload: any) => handleCall(getLastN, payload),
  // eslint-disable-next-line jsdoc/require-jsdoc
  [FETCH_RANGE]: (payload: any) => handleCall(getRange, payload),
  // eslint-disable-next-line jsdoc/require-jsdoc
  [WIPE_DATABASE]: (payload: any) => handleCall(wipe, payload),
  // eslint-disable-next-line jsdoc/require-jsdoc
  [FETCH_METADATA]: (payload: any) => handleCall(getMetadata, payload)
}

/**
 * Reducer method for handling call based on given key
 *
 * @param {string} key network call key
 * @param {any} payload data payload for endpoint
 * @returns {Promise<HandledResponse>} response
 */
export const requestReducer = (key: string, payload: any): Promise<HandledResponse | void> => connections[key](payload)

/**
 * Method for calling a given network call function
 *
 * @param {Function} call endpoint function
 * @param {any} payload data payload for endpoint
 * @returns {Promise<HandledResponse>} response
 */
const handleCall = async (call: Function, payload: any): Promise<HandledResponse | void> => {
  const response = await call(payload)
  return handleResponse(response, payload)
}

const NETWORK_ERROR_MESSAGE = { message: 'Network issue, try again later' }
const CANCELED_ERROR_MESSAGE = { message: 'Request canceled by clean-up' }

/**
 * Method for handling response status
 *
 * @param {NetworkResponse} response response from network reducer
 * @param {any} payload payload for endpoint
 * @param {Function} payload.next callback for success
 * @param {Function} payload.error callback for error
 * @returns {Promise<HandledResponse>} response
 */
const handleResponse = (response: NetworkResponse, payload: { next: Function; error: Function }): HandledResponse | void => {
  if (checkCanceled(response)) return payload.next(CANCELED_ERROR_MESSAGE.message, 200)
  if (!checkNetworkError(response)) return payload.error(NETWORK_ERROR_MESSAGE, 500)
  if (!checkStatus(response.status)) return payload.error(response.data, response.status)
  return payload.next(response.data, response.status)
}

/**
 * Method for checking if call was successful or failure
 *
 * @param {number} status HTTP response code
 * @returns {boolean} status
 */
const checkStatus = (status: number): boolean => {
  return status >= 200 && status <= 299
}

/**
 * Method for checking if call was cancelled or aborted
 *
 * @param {NetworkResponse | string} response response to be checked
 * @returns {boolean} cancelled
 */
const checkCanceled = (response: NetworkResponse | string): boolean => {
  return response === 'canceled' || typeof response === 'string'
}

/**
 * Method for checking if network error has occurred
 *
 * @param {NetworkResponse} response response to be checked
 * @returns {boolean} error
 */
const checkNetworkError = (response: any): boolean => {
  if (!response) return false
  return 'data' in response
}
