import { useToast } from '@chakra-ui/react'
import { ReactNode } from 'react'
import { ToastStatus } from '../constants'

interface Exports {
  errorToast: Function
  successToast: Function
  pendingToast: Function
  resolveToast: Function
  updateToast: Function
  warningToast: Function
  infoToast: Function
}

/**
 * Functional Handler Component, for handling all toast related actions
 *
 * @returns {Exports} exported functions
 */
const ToastHandler = (): Exports => {
  const DURATION = 6e3
  const POSITION = 'bottom-left'
  const toast = useToast()
  const TOAST_ID = 'pending-toast'

  /**
   * Method for launching a Error Toast, also closes all pending toasts
   *
   * @param {string} title toast title
   * @param {string} description toast description
   */
  const errorToast = (title: string, description: string): void => {
    toast.close(TOAST_ID)
    toast({ status: ToastStatus.Error, duration: null, title, description, isClosable: true, position: POSITION })
  }
  /**
   * Method for launching a Error Toast
   *
   * @param {string} title toast title
   * @param {string} description toast description
   */
  const warningToast = (title: string, description: string): void => {
    toast.closeAll()
    toast({ status: ToastStatus.Warning, duration: DURATION, title, description, isClosable: true, position: POSITION })
  }
  /**
   * Method for launching a Error Toast
   *
   * @param {string} title toast title
   * @param {string} description toast description
   */
  const infoToast = (title: string, description: string): void => {
    toast.closeAll()
    toast({ status: ToastStatus.Info, duration: DURATION, title, description, isClosable: true, position: POSITION })
  }
  /**
   * Method for launching a Error Toast
   *
   * @param {string} title toast title
   * @param {string} description toast description
   */
  const successToast = (title: string, description: string): void => {
    toast({ status: ToastStatus.Success, duration: DURATION, title, description, isClosable: true, position: POSITION })
  }
  /**
   * Method for launching a Error Toast
   *
   * @param {string} title toast title
   * @param {string} description toast description
   */
  const pendingToast = (title: string, description: string): void => {
    toast.closeAll()
    toast({ id: TOAST_ID, status: ToastStatus.Info, duration: null, title, description, isClosable: false, position: POSITION })
  }

  /**
   * Method for updating a pending toast
   *
   * @param {ToastStatus} status new toast status
   * @param {string} title new toast title
   * @param {string} description new toast description
   */
  const updateToast = (status: ToastStatus, title: string, description: ReactNode | string): void => {
    toast.update(TOAST_ID, {
      status,
      duration: null,
      title,
      description,
      isClosable: false,
      position: POSITION
    })
  }

  /**
   * Method for resolving a pending toast
   *
   * @param {ToastStatus} status new toast status
   * @param {string} title new toast title
   * @param {string} description new toast description
   */
  const resolveToast = (status: ToastStatus, title: string, description: string): void => {
    toast.update(TOAST_ID, { status, duration: status === ToastStatus.Error ? null : DURATION, title, description, isClosable: true, position: POSITION })
  }

  return { errorToast, successToast, pendingToast, resolveToast, updateToast, warningToast, infoToast }
}

export default ToastHandler
