import { useState, useEffect, Dispatch, SetStateAction } from 'react'

/**
 * Accessor for stored value, used when re-accessing the variable on new load
 *
 * @template T
 * @param {string} key accessor for this stored value
 * @param {T} defaultValue default value
 * @returns {T} stored value
 */
function getStorageValue<T> (key: string, defaultValue: T): T {
  // getting stored value
  let saved: string | null = null
  if (typeof window !== 'undefined') saved = localStorage.getItem(key)
  if (saved?.toString() === 'undefined') return defaultValue
  const initial = JSON.parse(saved!)
  return initial || defaultValue
}

/**
 * Custom React hook, for saving the useState with local-storage in the window
 *
 * @template T
 * @param {string} key accessor for this stored value
 * @param {T} defaultValue default value
 * @returns {[T, Dispatch<SetStateAction<T>>]} useState-like variable + function
 */
export function useLocalStorage<T> (key: string, defaultValue: T): [T, Dispatch<SetStateAction<T>>] {
  const [value, setValue] = useState<T>(() => {
    return getStorageValue<T>(key, defaultValue)
  })

  // Listener for changes
  useEffect((): any => {
    if (!value && typeof window !== 'undefined') return window.localStorage.setItem(key, JSON.stringify(''))
    if (typeof window !== 'undefined') localStorage.setItem(key, JSON.stringify(value))
  }, [key, value])

  return [value, setValue]
}
