import { Signal } from '../globals'
import { ToastStatus } from './constants'

export const NAMEVALIDATOR = /^[\w_-]{1,64}$/i

/**
 * Capitalizes chart at index 0 of given string
 *
 * @param {string} str string argument
 * @returns {string} handled string
 */
export const capitalize = (str: string): string => {
  return str.charAt(0).toUpperCase() + str.slice(1)
}

/**
 * Normalizes a string with given "_" chars to seperate words with space
 *
 * @param {string} name string
 * @returns {string} handled string
 */
export const normalizeName = (name: string): string => {
  const words = name.split('-')
  if (words.length < 2) return capitalize(name)
  let _name = ''
  words.forEach((word: string) => (_name += capitalize(word) + ' '))
  return _name
}

/**
 * Normalizes a number to string of 2 digits
 *
 * @param {number} number number
 * @returns {string} string
 */
export const normalizeNumber = (number: number): string => {
  return number.toString().length === 1 ? `0${number}` : number.toString()
}

/**
 * Acts as ratelimit for any given callback function
 *
 * @param {Function} callback any given function
 * @param {number} delay wait time befor callback execution
 * @returns {Function} debounce function
 */
export const debounce = (callback: Function, delay: number = 1000): Function => {
  let timeout: any

  return (...args: any) => {
    clearTimeout(timeout)
    // eslint-disable-next-line n/no-callback-literal
    timeout = setTimeout(() => callback(...args), delay)
  }
}

/**
 * Throttle function gives a buffer persiod after callback is called
 *
 * @param {Function} callback any given function
 * @param {number} delay wait time after callback execution
 * @returns {Function} throttle function
 */
export const throttle = (callback: Function, delay: number = 1000) => {
  let isWait = false

  return (...args: any) => {
    if (isWait) return
    // eslint-disable-next-line n/no-callback-literal
    callback(...args)
    isWait = true

    setTimeout(() => {
      isWait = false
    }, delay)
  }
}

/**
 * gets string version of hour minute into nubers
 *
 * @param {string} time time of day
 * @returns {object} hour minute numbers
 */
export const getHourMinute = (time: string): { h: number; m: number } => {
  let h = 0
  let m = 0
  try {
    h = parseInt(time.slice(0, 2))
    m = parseInt(time.slice(2, 4))

    return { h, m }
  } catch (err: any) {
    console.warn(err)
    return { h, m }
  }
}

/**
 * Handler for formatting european style time in string format
 *
 * @param {any} val value to be formatted
 * @returns {string} handled time of day
 */
export const format = (val: any): string => {
  const str: any = val.toString()
  if (str.length > 4) return ''
  if (str.length > 2) return str.slice(0, 2) + ':' + str.slice(2, str.length)
  return val
}

/**
 *  Handler for parsing a value into number or string-number
 *
 * @param {any} val value to be parsed
 * @returns {number | string} handled time of day
 */
export const parse = (val: string): number | string => {
  try {
    const result = parseInt(val.replace(/^\$/, ''))
    if (isNaN(result)) return ''
    return result
  } catch (err: any) {
    console.warn(err)
    return ''
  }
}

/**
 * Formats unix date into EN-US format
 *
 * @param {number} time unix timestamp
 * @returns {string} handled time
 */
export const getFormattedDate = (time: number): string => {
  try {
    const options: any = { weekday: 'long', year: 'numeric', month: 'long', day: 'numeric' }
    return new Intl.DateTimeFormat('en-US', options).format(new Date(time))
  } catch (err: any) {
    return ''
  }
}

/**
 * Sorting a map of doubles into list
 *
 * @param {Map<string, string>} double map of doubles
 * @returns {string[]} sorted doubles
 */
export const sortDoubles = (double: { [key: string]: number }): string[] => {
  try {
    const temp = Object.keys(double)
    return temp.sort()
  } catch (err: any) {
    console.warn(err)
    return []
  }
}

export const ERROR_PREFIX = 'Error when requesting: '
export const WARNING_PREFIX = 'Could not complete request: '
export const INFO_PREFIX = 'Info: '

/**
 * Toast Prefix generator
 *
 * @param {ToastStatus} status status of toast
 * @returns {string | null} prefix
 */
const prefix = (status: ToastStatus): string | null => {
  switch (status) {
    case ToastStatus.Error:
      return ERROR_PREFIX
    case ToastStatus.Warning:
      return WARNING_PREFIX
    case ToastStatus.Info:
      return INFO_PREFIX
    default:
      return null
  }
}

/**
 * Toast Sufix generator
 *
 * @param {string} _key constant passed in network call
 * @returns {string} sufix
 */
const sufix = (_key: string): string => {
  const result: string[] = []
  const items = _key.split('_')

  items.forEach((item: string) => {
    result.push(capitalize(item.toLowerCase()))
  })

  return result.toString().replaceAll(',', ' ')
}

/**
 * Gets title of Toast based on inputs
 *
 * @param {ToastStatus} status status of toast
 * @param {string} key constant passed in network call
 * @returns {string} title
 */
export const titleOf = (status: ToastStatus, key: string): string => {
  return `${prefix(status)}${sufix(key)}`
}

/**
 * Gets difference between to unix timestamps
 *
 * @param {number} start unix start
 * @param {number} end unix end
 * @returns {string} difference message
 */
export const getMinuteDifference = (start: number, end: number): string => {
  const startTime = new Date(start)
  const endTime = new Date(end)
  const difference = endTime.getTime() - startTime.getTime() // This will give difference in milliseconds
  return 'Difference: ' + Math.round(difference / 60000) + ' minutes'
}

/**
 * Helper Method for reducing list of Signals to list of signal names
 *
 * @param {Signal[]} signals input
 * @returns {string[]} names
 */
export const getSignalNames = (signals: Signal[]): string[] => {
  return signals.reduce((names: string[], signal: Signal) => {
    return names.concat(signal.name)
  }, [])
}
