import { Changes } from '../../proto/cdplogreader'
import healthCheckMiddleware, { HCFail } from '../../services/cache/healthCheckMiddleware'
import MqttPool from '../../services/mqtt/mqttPool'

/**
 * Handles a last n request
 *
 * @param {string} vessel The vessel to request from
 * @param {string[]} variables The variables to request
 * @param {number} lastn The number of changes to request
 * @returns {Changes & {bytetransfer: number}} The response from the cache and the amount of bytes transfered
 */
const requestLastN = async (vessel: string, variables: string[], lastn: number): Promise<Changes & {bytetransfer: number}> => {
  const pubTopic = `wlcache/data/${vessel}/request`
  const subTopic = `wlcache/data/${vessel}/response`

  const client = await MqttPool.get()

  /**
   * Executes and handles the request
   */
  const execute = async () => {
    return await healthCheckMiddleware<Changes & {bytetransfer: number}>(client, (resolve, reject) => {
      client.subscribe(subTopic)

      client.on('message', (topic, payload) => {
        if (topic !== subTopic) return
        const changes = Changes.decode(payload) as Changes & {bytetransfer: number}
        if (changes.lastN?.nChanges !== lastn.toString()) return
        changes.bytetransfer = payload.length
        resolve(changes)
      })

      client.publish(pubTopic, Buffer.from(Changes.encode({
        lastN: { nChanges: lastn + '' },
        signal: variables.map(d => {
          return {
            name: d,
            double: {}
          }
        }),
        fromTo: undefined
      }).finish()))
    })
  }

  try {
    return await execute()
  } catch (e: any) {
    if (e === HCFail) {
      throw new Error(HCFail)
    } else {
      throw e
    }
  } finally {
    MqttPool.release(client)
  }
}
export default requestLastN
