import { Changes } from '../../proto/cdplogreader'
import healthCheckMiddleware, { HCFail } from '../../services/cache/healthCheckMiddleware'
import MqttPool from '../../services/mqtt/mqttPool'
import { performance } from 'perf_hooks'
import { Worker } from 'worker_threads'
import path from 'path'

/**
 * Handler for "request range" requests
 *
 * @param {string} vessel The vessel to request from
 * @param {string[]} variables The variables to request
 * @param {number} from The start of the range to request
 * @param {number} to The end of the range to request
 * @returns {Changes & {bytetransfer: number}} The response from the cache and the amount of bytes transfered
 */
const requestRange = async (vessel: string, variables: string[], from: number, to: number): Promise<Changes & {bytetransfer: number}> => {
  const pubTopic = `wlcache/data/${vessel}/request`
  const subTopic = `wlcache/data/${vessel}/response`

  const client = await MqttPool.get()

  /**
   * Executes a Cache API call wrapped in the healthcheck middleware
   */
  const execute = async () => {
    return await healthCheckMiddleware<Changes & {bytetransfer: number}>(client, (resolve, reject) => {
      try {
        client.subscribe(subTopic)

        /**
         * Handles a response
         *
         * @param {string} topic The response topic
         * @param {Buffer} payload The response payload
         */
        const listener = async (topic: string, payload: Buffer) => {
          if (topic !== subTopic) return
          const se = performance.now()
          const changes = await new Promise<Changes & {bytetransfer: number}>(resolve => {
            const w = new Worker(path.resolve(process.cwd(), './dst/services/workers/decodeChanges.js'), { workerData: payload })
            w.on('message', buffer => {
              resolve(buffer)
            })
          })
          const ee = performance.now()
          console.log('Decoded in ', (ee - se))
          if (changes.fromTo?.fromTime !== from.toString() || changes.fromTo.toTime !== to.toString()) return
          changes.bytetransfer = payload.length
          resolve(changes)
        }

        client.addListener('message', listener)
        client.publish(pubTopic, Buffer.from(Changes.encode({
          fromTo: {
            fromTime: from + '',
            toTime: to + ''
          },
          signal: variables.map(variable => { return { name: variable, double: {} } }),
          lastN: undefined
        }).finish()))
      } catch (e) {
        reject(e)
      }
    })
  }

  try {
    return await execute()
  } catch (e: any) {
    if (e === HCFail) {
      throw new Error(HCFail)
    } else {
      throw e
    }
  } finally {
    MqttPool.release(client)
  }
}
export default requestRange
