import { Changes } from '../../proto/cdplogreader'
import healthCheckMiddleware, { HCFail } from '../../services/cache/healthCheckMiddleware'
import mqtt from 'mqtt'

/**
 * Request range handler modified for live requests
 *
 * @param {string} vessel The vessel to request from
 * @param {string[]} variables The variables to request
 * @param {number} from The start of the period to request
 * @param {number} to The end of the period to request
 * @param {mqtt.MqttClient} client The mqtt client to use for the request
 * @returns {Promise<Changes>} The response from the cache
 */
const requestRange = async (vessel: string, variables: string[], from: number, to: number, client: mqtt.MqttClient): Promise<Changes> => {
  const pubTopic = `wlcache/data/${vessel}/request`
  const subTopic = `wlcache/data/${vessel}/response`

  /**
   * Executes the range request
   */
  const execute = async () => {
    return await healthCheckMiddleware<Changes>(client, (resolve, reject) => {
      try {
        client.subscribe(subTopic)

        /**
         * Handles incoming response
         *
         * @param {string} topic The response topic
         * @param {Buffer} payload The response payload
         */
        const listener = (topic: string, payload: Buffer) => {
          if (topic === subTopic) {
            const changes = Changes.decode(payload)
            if (changes.fromTo?.fromTime !== from.toString() || changes.fromTo.toTime !== to.toString()) return
            resolve(changes)
            client.removeListener('message', listener)
          }
        }
        client.addListener('message', listener)
        client.publish(pubTopic,
          Buffer.from(Changes.encode({
            fromTo: { fromTime: from + '', toTime: to + '' },
            lastN: undefined,
            signal: variables.map(d => {
              return {
                name: d,
                double: {}
              }
            })
          }).finish()))
      } catch (e) {
        reject(e)
      }
    })
  }

  try {
    return await execute()
  } catch (e: any) {
    if (e === HCFail) {
      throw new Error(HCFail)
    } else {
      throw e
    }
  }
}
export default requestRange
