/* eslint no-unmodified-loop-condition: "off" */

import express from 'express'
import cors, { CorsOptions } from 'cors'
import WebSocket from 'ws'
import getMQTT from './services/mqtt/getMQTT'
import requestRangeLive from './handlers/requests/requestRangeLive'
import vesselid from './routes/vessel/vessel'
import commands from './routes/commands/commands'

/**
 *
 */
const runServer = () => {
  const app = express()
  const expSrv = app.listen(process.env.PORT || 3000, () => { console.log(`Listening on ${process.env.PORT || 3000}`) })
  app.use(express.json())

  const corsOptions: CorsOptions = {
    origin: /.*/,
    methods: ['GET', 'POST']
  }

  app.use(cors(corsOptions))

  const ws = new WebSocket.Server({
    noServer: true,
    path: '/live'
  })

  ws.on('connection', async (socket, request) => {
    const urlComponents = request.url?.split('?')!
    if (urlComponents == null) socket.close(0, 'Missing path parameters')
    const query = urlComponents[1].split('&').map(q => q.split('='))
    const vessel = query.find(q => q[0] === 'vessel')?.[1]
    const dataTypes = query.find(q => q[0] === 'data')?.[1]?.split(',')

    if (dataTypes == null) {
      socket.send('Missing data query parameter')
      socket.terminate()
    }
    if (vessel == null) {
      socket.send('Missing vessel query parameter')
      socket.terminate()
    }

    let running = true
    socket.on('close', () => {
      running = false
    })
    const mqttClient = await getMQTT()

    let fromTime = Date.now() - 1e4

    /*
    While the websocket is up and running, perform requests in one second intervals.
    The requests themselves are handled independently of the loop, so the loop always
    sends a new request after 1 second, regardless of the previous requests status
     */
    while (running) {
      await sleep(1e3)
      requestRangeLive(vessel!, dataTypes!, fromTime, fromTime + 1e3, mqttClient)
        .then(signals => {
          socket.send(JSON.stringify(signals))
        }).catch(e => {
          console.log(e)
        })
      fromTime += 1e3
    }
  })

  expSrv.on('upgrade', (req, socket, head) => {
    ws.handleUpgrade(req, socket, head, (websocket, request) => {
      ws.emit('connection', websocket, request)
    })
  })

  app.use('/command', commands)
  app.use('/:vessel', vesselid)
  app.get('/', (req, res) => {
    res.send('Hello 😊')
  })
}

export default runServer

/**
 * Creates promise that sleeps for the given length
 *
 * @param {number} ms The number of milliseconds to sleep
 */
const sleep = (ms: number): Promise<void> => {
  return new Promise<void>((resolve) => {
    setTimeout(() => {
      resolve()
    }, ms)
  })
}
