import mqtt from 'mqtt'
import getMQTT from '../mqtt/getMQTT'

/**
 * Class representing a HealthCheck SideCar observer
 * Uniquely ID-ed and must be kept reference to if used
 */
export class HealthCheckSideCarObserver {
  /**
   * This observers call function
   *
   * @private
   */
  private readonly observer: () => void

  /**
   * This observers unique key
   *
   * @private
   */
  private readonly key: number

  /**
   * The amount of pongs this observer has accumulated
   *
   * @private
   */
  private accumulatedPongs: number = 0

  /**
   * Is this observer canceled?
   *
   * @private
   */
  private canceled: boolean = false

  /**
   * HealthCheckSideCarObserver constructor
   *
   * @param {() => void} observer The observers call function
   * @param {number} key The observers unique id
   */
  constructor (observer: () => void, key: number) {
    this.observer = observer
    this.key = key
  }

  /**
   * Calls and cancels observer
   */
  public call () {
    this.observer()
    this.cancel()
  }

  /**
   * Returns the observers unique key
   *
   * @returns {number} The observers unique key
   */
  public getKey () {
    return this.key
  }

  /**
   * Updates the observers accumulated pongs
   * If the observer has passed 50 pongs it will self-call and cancel
   * 50 pongs means approximately 5 minutes, which is an unlikely slow response time by the cache
   * This number is arbitrary and has been set based on typical response times with a good margin
   */
  public pong () {
    this.accumulatedPongs++
    if (this.accumulatedPongs > 50) {
      HealthCheckSideCar.cancel(this)
      this.call()
    }
  }

  /**
   * Sets teh observers canceled property to true
   */
  public cancel () {
    this.canceled = true
  }

  /**
   * Getter for observers canceled property
   *
   * @returns {boolean} True if this observer has been called or canceled
   */
  public isCanceled (): boolean {
    return this.canceled
  }
}

/**
 * Class that can be observed to get notified whenever the cache, or mqtt broker is down or unresponsive
 */
export class HealthCheckSideCar {
  /**
   * The mqtt client the sidecar uses to subscribe to healthcheck pongs
   *
   * @private
   */
  private static client: mqtt.MqttClient

  /**
   * The sidecar interval which checks for pongs
   *
   * @private
   */
  private static interval: NodeJS.Timer

  /**
   * Boolean to keep track of whether or not a pong has been received
   *
   * @private
   */
  private static ponged: boolean = true

  /**
   * A map of all observers by their keys
   *
   * @private
   */
  private static observers: Record<number, HealthCheckSideCarObserver> = {}

  /**
   * The current key index, a number between 0 and 19
   * The value 19 comes from the Mqtt pools size
   *
   * @private
   */
  private static keyIndex: number = 0

  /**
   * Initialize the healthcheck sidecar
   */
  public static async init () {
    await this.makeClient(false)

    // Check for "pongs" every 6 seconds (pong-rate is 3 seconds from the cache)
    this.interval = setInterval(() => {
      if (!this.ponged) {
        // If no pongs have been received, call the observers to notify them that the healthcheck has failed
        const observers = this.observers
        this.observers = {}
        Object.values(observers).forEach(observer => {
          observer.call()
        })
      }

      // Notify observers that a 6-second interval has passed
      Object.values(this.observers).forEach(observer => observer.pong())
      this.ponged = false
    }, 6e3)
  }

  /**
   * Creates an MQTT client which will perform the healthchecks
   *
   * @param {boolean} error Whether or not this client was made to recover from an error
   */
  private static async makeClient (error: boolean) {
    if (error) console.log('Attempting to reconnect healthcheck sidecar')

    try {
      // Create new mqtt client and set it up to deal with healthcheck "pongs"
      this.client = await getMQTT(() => {
        console.log('Healthcheck sidecar up')
      })
      this.client.subscribe('wlcache/healthcheck')
      this.client.on('message', topic => {
        if (topic !== 'wlcache/healthcheck') return
        this.ponged = true
      })

      // If an error is thrown by the client, kill it and create a new one
      this.client.on('error', async (e) => {
        console.log('Healthcheck sidecar down')
        this.client.end()
        await this.makeClient(true)
      })
    } catch (e) {
      // If the client throws an error in the creation stage, wait 5 seconds and try again
      // This usually means that the mqtt broker is not available
      await new Promise<void>(resolve => {
        setTimeout(() => resolve(), 5e3)
      })
      await this.makeClient(true)
    }
  }

  /**
   * Add an observer to the healthcheck sidecar
   *
   * @param {() => void} func the function the observer should call on failure
   * @returns {HealthCheckSideCarObserver} A reference to the observer
   */
  public static observe (func: () => void): HealthCheckSideCarObserver {
    // Avoid this observer colliding with another
    while (this.observers[this.keyIndex] != null) this.keyIndex = (this.keyIndex + 1) % 20

    const observer = new HealthCheckSideCarObserver(func, this.keyIndex)
    this.keyIndex = (this.keyIndex + 1) % 20
    this.observers[observer.getKey()] = observer
    return observer
  }

  /**
   * Cancel an observer by removing it from the list of observers.
   * This must be called if an observer is no longer to be called on a healthcheck failure
   *
   * @param {HealthCheckSideCarObserver} observer The observer to cancel
   */
  public static cancel (observer: HealthCheckSideCarObserver) {
    // If it is already canceled return
    if (observer.isCanceled()) return

    // Update the observers internal state then remove it from observers
    observer.cancel()
    delete this.observers[observer.getKey()]
  }
}
