import mqtt from 'mqtt'
import { clearTimeout } from 'timers'

/**
 * The default callback when getMQTT succedes
 */
const defaultCB = () => {
  console.log('MQTT connected')
}

/**
 * The default callback when getMQTT fails
 *
 * @param {any} err The reason or error
 */
const defaultOnErr = (err?: any) => {
  if (err) {
    console.log(err)
  } else {
    console.log('Could not connect to MQTT broker..')
  }
}

/**
 * Creates and connects pre-configured MQTT client
 *
 * @param {() => void} onConnectCB The callback to run on successful connectoin
 * @param {(err: any) => void} onErr The callback to return if connection fails
 * @returns {Promise<mqtt.MqttClient>} A preconfigured mqtt client
 */
const getMQTT = (onConnectCB: () => void = defaultCB, onErr: (err?: any) => void = defaultOnErr): Promise<mqtt.MqttClient> => {
  return new Promise<mqtt.MqttClient>((resolve, reject) => {
    try {
      const mqttClient = mqtt.connect('', {
        host: process.env.MQTT_BROKER_ADDRESS,
        protocol: process.env.MQTT_REQUIRE_TLS !== 'false' ? 'mqtts' : 'mqtt',
        username: process.env.MQTT_USER ?? undefined,
        password: process.env.MQTT_PWD ?? undefined,
        port: process.env.MQTT_PORT != null ? parseInt(process.env.MQTT_PORT) : process.env.MQTT_REQUIRE_TLS !== 'false' ? 8883 : 1883,
        connectTimeout: 5e3,
        reconnectPeriod: 1e3,
        keepalive: 30
      })
      const timeout = setTimeout(() => {
        mqttClient.emit('error', 'Cannot reach broker')
      }, 1e4)

      mqttClient.on('connect', async () => {
        clearTimeout(timeout)
        onConnectCB()
        resolve(mqttClient)
      })

      mqttClient.on('error', err => {
        reject(err)
        mqttClient.end()
      })
    } catch (e) {
      reject(onErr(e))
    }
  })
}
export default getMQTT
